import warnings

import numpy as np

from .typing import Cancer500Nodule, Comment, Review, Texture


def get_nodules(protocol, series_number, slice_locations):
    if protocol['nodules'] is None:
        num_doctors = len(protocol['doctors'])
        assert num_doctors in [3, 6]

        if len([d for d in protocol['doctors'] if definetely_no_nodules(d['comment'])]) > num_doctors / 2:
            return []
        else:
            raise ValueError

    assert protocol['nodules']

    nodules = []
    for nodule in protocol['nodules']:
        annotations = dict(get_nodule_annotations(nodule[-1], series_number, slice_locations))
        if not annotations:
            raise ValueError

        nodules.append(annotations)

    return nodules


def definetely_no_nodules(overall_comment):
    overall_comment = overall_comment.lower()
    prefixes = ['нет очагов', 'очагов нет', 'очаги не выявлены', 'достоверно очагов нет']
    return any(overall_comment.startswith(p) for p in prefixes)


def get_nodule_annotations(nodule: dict, series_number: int, slice_locations: list):
    for rater, ann in nodule.items():
        if ann is None:
            continue

        if 'series_no' in ann and str(series_number) not in ann['series_no']:
            warnings.warn('Cannot check that annotation belongs to this particular series.')
            continue

        try:
            yield rater, parse_nodule_annotation(ann, slice_locations)
        except ValueError as e:
            warnings.warn(str(e))
            continue


def parse_nodule_annotation(ann: dict, slice_locations: list):
    return Cancer500Nodule(
        center_voxel=parse_center_voxel(ann, slice_locations),
        review=parse_review(ann),
        comment=parse_comment(ann),
        diameter_mm=parse_diameter_mm(ann),
        texture=parse_texture(ann),
        malignancy=parse_malignancy(ann),
    )


def parse_center_voxel(ann: dict, slice_locations: list):
    i, j = int(ann['x']), int(ann['y'])
    assert i == ann['x']
    assert j == ann['y']

    assert 'z type' in ann
    assert ann['z type'].strip() == 'mm'
    diff = np.abs(np.array(slice_locations) - ann['z'])
    if np.min(diff) >= 1:
        raise ValueError('Cannot determine slice.')
    slc = np.argmin(diff)

    comments = [review['comment'] for review in ann['expert decision']]
    if 'z = 258 = -151,6 ' in comments:
        slc = 258
    elif 'не 134 а 143 по оси Х' in comments:
        i = 143
    elif (
        'неправильная координата х (должно быть 73, а не 734). сосуд, несовпадение типа (другое), неверный размер'
        in comments
    ):
        i = 73
    elif 'ошибка в координате Y - должно быть 296, тогда очаг есть' in comments:
        j = 296
    elif 'срез съехал на два ниже' in comments:
        slc -= 2
    elif set(comments) & {
        'очага нет',
        'промахно',
        'промахнулись с координатой х',
        'часть координат не совпадает с топикой очага',
        'часть координат не совпадает с топикой очага, неверный размер',
    }:
        raise ValueError('Cannot detetmine center voxel')

    return i, j, slc


def parse_review(ann: dict):
    decisions = {review['decision'] for review in ann['expert decision']}
    if 'confirmed' in decisions:
        return Review.Confirmed
    elif 'confirmed_partially' in decisions:
        return Review.ConfirmedPartially
    elif 'doubt' in decisions:
        return Review.Doubt
    elif 'rejected' in decisions:
        return Review.Rejected
    else:
        raise ValueError(decisions)


def parse_comment(ann: dict):
    comments = {review['comment'] for review in ann['expert decision']}
    if 'кальцинат, несовпадение типа (другое)' in comments:
        return Comment.Calcium
    elif 'фиброз' in comments:
        return Comment.Fibrosis
    elif 'внутрилегочный л\\у' in comments:
        return Comment.LymphNode
    elif 'очаг с кальцинацией, несовпадение типа (другое)' in comments:
        return Comment.Calcified
    elif 'бронхоэктаз с содержимым, несовпадение типа (другое)' in comments:
        return Comment.Bronchiectasis
    elif 'сосуд' in comments:
        return Comment.Vessel


def parse_diameter_mm(ann: dict):
    if any('неверный размер' in review['comment'].lower() for review in ann['expert decision']):
        return

    return round(ann['diameter (mm)'], 2)


def parse_texture(ann: dict):
    nodule_types = {review['type'] for review in ann['expert decision']} & {'#0S', '#1PS', '#2GG', 'другое'}
    if nodule_types:
        assert len(nodule_types) == 1
        (nodule_type,) = nodule_types
    elif parse_review(ann) in [Review.Confirmed, Review.ConfirmedPartially, Review.Doubt]:
        assert ann['type'] in ['#0S', '#1PS', '#2GG']
        nodule_type = ann['type']
    else:
        return

    if nodule_type == '#0S':
        return Texture.Solid
    elif nodule_type == '#1PS':
        return Texture.PartSolid
    elif nodule_type == '#2GG':
        return Texture.GroundGlass
    elif nodule_type == 'другое':
        return Texture.Other


def parse_malignancy(ann: dict):
    malignant = [review['malignant'] for review in ann['expert decision']]
    if all(malignant):
        return True
    elif not any(malignant):
        return False



import contextlib
from pathlib import Path
from typing import Union
from zipfile import ZipFile

import nibabel
import numpy as np
import pandas as pd

from .internals import Dataset, field, licenses, register
from .utils import open_nii_gz_file, unpack


@register(
    body_region='Head',
    license=licenses.CC_BYNCSA_40,
    link='http://www.braintumorsegmentation.org/',
    modality=('MRI T1', 'MRI T1Gd', 'MRI T2', 'MRI T2-FLAIR'),
    prep_data_size='8,96G',
    raw_data_size='15G',
    task=('Segmentation', 'Classification', 'Domain Adaptation'),
)
class BraTS2021(Dataset):
    """
    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Download links:
    2021: http://www.braintumorsegmentation.org/

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = BraTS2021(root='/path/to/archives/root')
    >>> print(len(ds.ids))
    # 5880
    >>> print(ds.image(ds.ids[0]).shape)
    # (240, 240, 155)

    References
    ----------
    """

    @property
    def ids(self):
        return sorted(_get_ids_or_file(self.root, 'TrainingData') + _get_ids_or_file(self.root, 'ValidationData'))

    @field
    def fold(self, i) -> str:
        return 'ValidationData' if _get_ids_or_file(self.root, 'ValidationData', check_id=i) else 'TrainingData'

    @property
    def mapping21_17(self) -> pd.DataFrame:
        return pd.read_csv(self.root / 'BraTS21-17_Mapping.csv')

    @field
    def subject_id(self, i) -> str:
        return i.rsplit('_', 1)[0]

    @field
    def modality(self, i) -> str:
        return i.rsplit('_', 1)[1]

    @field
    def image(self, i) -> np.ndarray:
        root, relative = _get_ids_or_file(self.root, self.fold(i), check_id=i, return_image=True)
        with _load_nibabel_probably_from_zip(root, relative, '.', '.zip') as nii_image:
            return np.asarray(nii_image.dataobj)

    def mask(self, i) -> Union[np.ndarray, None]:
        if self.fold(i) == 'ValidationData':
            return None
        else:
            root, relative = _get_ids_or_file(self.root, self.fold(i), check_id=i, return_segm=True)
            with _load_nibabel_probably_from_zip(root, relative, '.', '.zip') as nii_image:
                return np.asarray(nii_image.dataobj)

    def spacing(self, i):
        """Returns the voxel spacing along axes (x, y, z)."""
        root, relative = _get_ids_or_file(self.root, self.fold(i), check_id=i, return_image=True)
        with _load_nibabel_probably_from_zip(root, relative, '.', '.zip') as nii_image:
            return tuple(nii_image.header['pixdim'][1:4])

    @field
    def affine(self, i) -> np.ndarray:
        """Returns 4x4 matrix that gives the image's spatial orientation."""
        root, relative = _get_ids_or_file(self.root, self.fold(i), check_id=i, return_image=True)
        with _load_nibabel_probably_from_zip(root, relative, '.', '.zip') as nii_image:
            return nii_image.affine


def _get_ids_or_file(
    base_path,
    archive_name_part: str = 'TrainingData',
    check_id: str = None,
    return_image: bool = False,
    return_segm: bool = False,
):
    # TODO: implement the same functionality for folder extraction.
    ids = []
    for archive in base_path.glob('*.zip'):
        if archive_name_part in archive.name:
            with ZipFile(archive) as zf:
                for zipinfo in zf.infolist():
                    if not zipinfo.is_dir():
                        file = Path(zipinfo.filename)
                        _id = file.stem.replace('.nii', '')

                        if 'seg' not in _id:
                            ids.append(_id)

                        if (check_id is not None) and (check_id == _id):
                            if return_segm:
                                return str(archive), str(file)[: -len('.nii.gz')].rsplit('_', 1)[0] + '_seg.nii.gz'

                            if return_image:
                                return str(archive), str(file)

                            return True  # if check_id in archive

    return ids if (check_id is None) else False  # if check_id not in archive


@contextlib.contextmanager
def _load_nibabel_probably_from_zip(root: str, relative: str, archive_root_name: str = None, archive_ext: str = None):
    with unpack(root, relative, archive_root_name, archive_ext) as (unpacked, is_unpacked):
        if is_unpacked:
            yield nibabel.load(unpacked)
        else:
            with open_nii_gz_file(unpacked) as nii_image:
                yield nii_image



from typing import Sequence, Union

import numpy as np
from connectome import Transform
from imops import zoom

from ..utils import Numeric, propagate_none


class CanonicalMRIOrientation(Transform):
    __inherit__ = True

    def image(image):
        return np.transpose(image, (1, 0, 2))[..., ::-1]

    def spacing(spacing):
        return tuple(np.array(spacing)[[1, 0, 2]].tolist())

    @propagate_none
    def schwannoma(schwannoma):
        return np.transpose(schwannoma, (1, 0, 2))[..., ::-1]

    @propagate_none
    def cochlea(cochlea):
        return np.transpose(cochlea, (1, 0, 2))[..., ::-1]

    @propagate_none
    def meningioma(meningioma):
        return np.transpose(meningioma, (1, 0, 2))[..., ::-1]


class Rescale(Transform):
    __inherit__ = True

    _new_spacing: Union[Sequence[Numeric], Numeric]
    _order: int = 1

    def _spacing(spacing, _new_spacing):
        _new_spacing = np.broadcast_to(_new_spacing, len(spacing)).copy()
        _new_spacing[np.isnan(_new_spacing)] = np.array(spacing)[np.isnan(_new_spacing)]
        return tuple(_new_spacing.tolist())

    def _scale_factor(spacing, _spacing):
        return np.float32(spacing) / np.float32(_spacing)

    def spacing(_spacing):
        return _spacing

    def image(image, _scale_factor, _order):
        return zoom(image.astype(np.float32), _scale_factor, order=_order)

    @propagate_none
    def schwannoma(schwannoma, _scale_factor, _order):
        return zoom(schwannoma.astype(np.float32), _scale_factor, order=_order) > 0.5

    @propagate_none
    def cochlea(cochlea, _scale_factor, _order):
        return zoom(cochlea.astype(np.float32), _scale_factor, order=_order) > 0.5

    @propagate_none
    def meningioma(meningioma, _scale_factor, _order):
        return zoom(meningioma.astype(np.float32), _scale_factor, order=_order) > 0.5



from functools import cached_property

import nibabel
import numpy as np

from ..internals import Dataset, licenses, register


@register(
    body_region='Chest',
    license=licenses.CC_BYNC_40,
    link='https://ribfrac.grand-challenge.org',
    modality='CT',
    raw_data_size='77.8 G',
    task='Segmentation',
)
class RibFrac(Dataset):
    """
    RibFrac dataset is a benchmark for developping algorithms on rib fracture detection,
    segmentation and classification. We hope this large-scale dataset could facilitate
    both clinical research for automatic rib fracture detection and diagnoses,
    and engineering research for 3D detection, segmentation and classification.


    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.


    Notes
    -----
    Data downloaded from here:
    https://doi.org/10.5281/zenodo.3893507 -- train Part1 (300 images)
    https://doi.org/10.5281/zenodo.3893497 -- train Part2 (120 images)
    https://doi.org/10.5281/zenodo.3893495 -- val (80 images)
    https://zenodo.org/record/3993380 -- test (160 images without annotation)



    References
    ----------
    Jiancheng Yang, Liang Jin, Bingbing Ni, & Ming Li. (2020).
    RibFrac Dataset: A Benchmark for Rib Fracture Detection,
    Segmentation and Classification
    """

    @property
    def ids(self):
        result = set()
        for folder in ['Part1', 'Part2', 'ribfrac-val-images', 'ribfrac-test-images']:
            result |= {v.name.split('-')[0] for v in (self.root / folder).iterdir()}

        return tuple(sorted(result))

    @cached_property
    def _id2folder(self):
        folders = [item for item in self.root.iterdir() if item.is_dir()]
        result_dict = {}
        for folder in folders:
            p = self.root / folder
            folder_ids = [v.name.split('-')[0] for v in p.iterdir()]
            folder_dict = {_id: p for _id in folder_ids}
            result_dict = {**result_dict, **folder_dict}

        return result_dict

    def image(self, i):
        image_path = self._id2folder[i] / f'{i}-image.nii.gz'
        image = nibabel.load(image_path).get_fdata()
        return image.astype(np.int16)

    def label(self, i):
        folder_path = self._id2folder[i]
        folder = folder_path.name
        if folder != 'ribfrac-test-images':
            if folder.startswith('Part'):
                label_path = folder_path / f'{i}-label.nii.gz'
            elif folder == 'ribfrac-val-images':
                dir = folder_path.parent / 'ribfrac-val-labels'
                label_path = dir / f'{i}-label.nii.gz'

            label = nibabel.load(label_path).get_fdata()
            return label.astype(np.int16)

    def affine(self, i):
        """The 4x4 matrix that gives the image's spatial orientation"""
        image_path = self._id2folder[i] / f'{i}-image.nii.gz'
        return nibabel.load(image_path).affine



import contextlib
import zipfile
from pathlib import Path

import pandas as pd

from ..internals.dataset import register_field


def csv_field(name, cast):
    def _loader(self, i):
        value = self._meta[i].get(name)
        if pd.isnull(value):
            return None
        if cast is not None:
            return cast(value)
        return value

    register_field('RSNABreastCancer', name, _loader)
    return _loader


@contextlib.contextmanager
def unpack(root: str, relative: str):
    unpacked = Path(root) / relative

    if unpacked.exists():
        yield unpacked, True
    else:
        with zipfile.Path(root, relative).open('rb') as unpacked:
            yield unpacked, False



from connectome.cache import unstable_module

from .__version__ import __version__
from .amos import AMOS
from .bimcv import BIMCVCovid19
from .brats2021 import BraTS2021
from .cancer_500 import MoscowCancer500
from .cc359 import CC359
from .cl_detection import CLDetection2023
from .covid_1110 import MoscowCovid1110
from .crlm import CRLM
from .crossmoda import CrossMoDA
from .ct_ich import CT_ICH
from .curvas import CURVAS
from .deeplesion import DeepLesion
from .egd import EGD
from .flare2022 import FLARE2022
from .hcp import HCP
from .lidc import LIDC
from .lits import LiTS
from .liver_medseg import LiverMedseg
from .medseg9 import Medseg9
from .midrc import MIDRC
from .mood import MOOD
from .mslub import MSLUB
from .nlst import NLST
from .nsclc import NSCLC
from .ribfrac import RibFrac
from .rsna_bc import RSNABreastCancer
from .stanford_coca import StanfordCoCa
from .totalsegmentator import Totalsegmentator
from .upenn_gbm import UPENN_GBM
from .verse import VerSe
from .vs_seg import VSSEG


unstable_module(__name__)



import nibabel as nb
import numpy as np
from deli import load

from .internals import Dataset, field as _field, register


@register(
    body_region='Head',
    license='EGD data license',
    link='https://xnat.bmia.nl/data/archive/projects/egd',
    modality=('FLAIR', 'MRI T1', 'MRI T1GD', 'MRI T2'),
    prep_data_size='107,49G',
    raw_data_size='40G',
    task='Segmentation',
)
class EGD(Dataset):
    """
    The Erasmus Glioma Database (EGD): Structural MRI scans, WHO 2016 subtypes,
    and segmentations of 774 patients with glioma [1]_.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    The access to the dataset could be requested at XNAT portal [https://xnat.bmia.nl/data/archive/projects/egd].

    To download the data in the compatible structure we recommend to use
    egd-downloader script [https://zenodo.org/record/4761089#.YtZpLtJBxhF].
    Please, refer to its README for further information.

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> egd = EGD(root='/path/to/downloaded/data/folder/')
    >>> print(len(egd.ids))
    # 774
    >>> print(egd.t1gd(egd.ids[215]).shape)
    # (197, 233, 189)
    >>> print(egd.manufacturer(egd.ids[444]))
    # Philips Medical Systems

    References
    ----------
    .. [1] van der Voort, Sebastian R., et al. "The Erasmus Glioma Database (EGD): Structural MRI scans,
           WHO 2016 subtypes, and segmentations of 774 patients with glioma."
           Data in brief 37 (2021): 107191.
           https://www.sciencedirect.com/science/article/pii/S2352340921004753

    """

    @property
    def ids(self):
        result = []
        for folder in (self.root / 'SUBJECTS').iterdir():
            for suffix in 'FLAIR', 'T1', 'T1GD', 'T2':
                result.append(f'{folder.name}-{suffix}')

        return tuple(sorted(result))

    @_field
    def brain_mask(self, i) -> np.ndarray:
        return nb.load(self.root / 'METADATA' / 'Brain_mask.nii.gz').get_fdata().astype(bool)

    @_field
    def deface_mask(self, i) -> np.ndarray:
        return nb.load(self.root / 'METADATA' / 'Deface_mask.nii.gz').get_fdata().astype(bool)

    def _image_file(self, i):
        i, suffix = i.rsplit('-', 1)
        return nb.load(self.root / 'SUBJECTS' / i / f'{suffix}.nii.gz')

    @_field
    def modality(self, i) -> str:
        _, suffix = i.rsplit('-', 1)
        return suffix

    @_field
    def subject_id(self, i) -> str:
        subject, _ = i.rsplit('-', 1)
        return subject

    @_field
    def affine(self, i) -> np.ndarray:
        return self._image_file(i).affine

    def spacing(self, i):
        # voxel spacing is [1, 1, 1] for all images in this dataset...
        return tuple(self._image_file(i).header['pixdim'][1:4])

    @_field
    def image(self, i) -> np.ndarray:
        # intensities are not integer-valued in this dataset...
        return np.asarray(self._image_file(i).dataobj)

    def _metadata(self, i):
        i, _ = i.rsplit('-', 1)
        return load(self.root / 'SUBJECTS' / i / 'metadata.json')

    @_field
    def genetic_and_histological_label_idh(self, i) -> str:
        return self._metadata(i)['Genetic_and_Histological_labels']['IDH']

    @_field
    def genetic_and_histological_label_1p19q(self, i) -> str:
        return self._metadata(i)['Genetic_and_Histological_labels']['1p19q']

    @_field
    def genetic_and_histological_label_grade(self, i) -> str:
        return self._metadata(i)['Genetic_and_Histological_labels']['Grade']

    @_field
    def age(self, i) -> float:
        return self._metadata(i)['Clinical_data']['Age']

    @_field
    def sex(self, i) -> str:
        return self._metadata(i)['Clinical_data']['Sex']

    @_field
    def observer(self, i) -> str:
        return self._metadata(i)['Segmentation_source']['Observer']

    @_field
    def original_scan(self, i) -> str:
        return self._metadata(i)['Segmentation_source']['Original scan']

    @_field
    def manufacturer(self, i) -> str:
        return self._metadata(i)['Scan_characteristics']['Manufacturer']

    @_field
    def system(self, i) -> str:
        return self._metadata(i)['Scan_characteristics']['System']

    @_field
    def field(self, i) -> str:
        return self._metadata(i)['Scan_characteristics']['Field']

    @_field
    def mask(self, i) -> np.ndarray:
        i, _ = i.rsplit('-', 1)
        return nb.load(self.root / 'SUBJECTS' / i / 'MASK.nii.gz').get_fdata().astype(bool)



import contextlib
import datetime
import functools
import itertools
import zipfile
from gzip import GzipFile
from os import PathLike
from pathlib import Path
from typing import List, Union

import nibabel
import numpy as np
from dicom_csv import get_common_tag, order_series, stack_images
from dicom_csv.exceptions import ConsistencyError, TagTypeError
from pydicom import Dataset, dcmread


Numeric = Union[float, int]
PathOrStr = Union[str, PathLike]


@contextlib.contextmanager
def unpack(root: PathOrStr, relative: str, archive_root_name: str = None, archive_ext: str = None):
    """Provides the absolute path to the file in both scenarios: inside archive or inside folder.

    Parameters
    ----------
    root : str, Path
        Absolute path to the downloaded archive or the unpacked archive root.
    relative : str, Path
        Relative file path inside the archive. Archive's root folder sholud be ommited.
    archive_root_name : str, Path, optional
        If `root` is a archive, it's root folder name shold be given.
    archive_ext: {'.zip'}, optional
        Compression algorithm used to create the archive

    Returns
    -------
    unpacked : Path
        Absolute file path to be opened.
    is_unpacked : {True, False}
        Reached file state. `True` if the file is located inside archive, `False` otherwise.
    """
    unpacked = Path(root) / relative

    if unpacked.exists():
        yield unpacked, True
    elif archive_ext == '.zip':
        with zipfile.Path(root, str(Path(archive_root_name, relative))).open('rb') as unpacked:
            yield unpacked, False
    else:
        raise ValueError('Unexpected file path or unsupported compression algorithm.')


@contextlib.contextmanager
def open_nii_gz_file(unpacked):
    """Opens ``.nii.gz`` file if it is packed in archive

    Examples
    --------
    >>> with unpack('/path/to/archive.zip', 'relative/file/path', 'root', '.zip') as (unpacked, is_unpacked):
    >>>     with open_nii_gz_file(unpacked) as image:
    >>>         print(np.asarray(image.dataobj).shape)
    # (512, 512, 256)
    """
    with GzipFile(fileobj=unpacked) as nii:
        nii = nibabel.FileHolder(fileobj=nii)
        yield nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})


def get_series_date(series):
    try:
        study_date = get_common_tag(series, 'StudyDate')
    except (TagTypeError, ConsistencyError):
        return

    if not isinstance(study_date, str) or not study_date.isnumeric() or len(study_date) != 8:
        return

    try:
        year = int(study_date[:4])
        month = int(study_date[4:6])
        day = int(study_date[6:])
    except TypeError:
        return

    if year < 1972:  # the year of creation of the first CT scanner
        return

    return datetime.date(year, month, day)


def propagate_none(func):
    @functools.wraps(func)
    def wrapper(x, *args, **kwargs):
        return None if (x is None) else func(x, *args, **kwargs)

    return wrapper


def deprecate(message=None):
    def decorator(func):
        return functools.wraps(func)(np.deprecate(message=message)(func))

    return decorator


def image_from_dicom_folder(folder: Union[str, Path]) -> np.ndarray:
    return stack_images(series_from_dicom_folder(folder))


def series_from_dicom_folder(folder: Union[str, Path]) -> List[Dataset]:
    return order_series([dcmread(p) for p in Path(folder).glob('*.dcm')])


# TODO: stolen from dpipe for now
def mask_to_box(mask: np.ndarray):
    """
    Find the smallest box that contains all true values of the ``mask``.
    """
    if not mask.any():
        raise ValueError('The mask is empty.')

    start, stop = [], []
    for ax in itertools.combinations(range(mask.ndim), mask.ndim - 1):
        nonzero = np.any(mask, axis=ax)
        if np.any(nonzero):
            left, right = np.where(nonzero)[0][[0, -1]]
        else:
            left, right = 0, 0
        start.insert(0, left)
        stop.insert(0, right + 1)
    return start, stop



from functools import cached_property
from zipfile import ZipFile

import nibabel
import numpy as np
import pandas as pd
from jboc import composed

from ..internals import Dataset, field, licenses, register
from ..utils import open_nii_gz_file, unpack


ARCHIVE_NAME_SEG = 'amos22.zip'
ARCHIVE_ROOT_NAME = 'amos22'
ERRORS = ['5514', '5437']  # these ids are damaged in the zip archives


# TODO: add MRI


@register(
    body_region='Abdomen',
    license=licenses.CC_BY_40,
    link='https://zenodo.org/record/7262581',
    modality=('CT', 'MRI'),
    raw_data_size='23G',  # TODO: update size with unlabelled
    prep_data_size='89,5G',
    task='Supervised multi-modality abdominal multi-organ segmentation',
)
class AMOS(Dataset):
    """
    AMOS provides 500 CT and 100 MRI scans collected from multi-center, multi-vendor, multi-modality, multi-phase,
    multi-disease patients, each with voxel-level annotations of 15 abdominal organs, providing challenging examples
    and test-bed for studying robust segmentation algorithms under diverse targets and scenarios. [1]

    Parameters
    ----------
    root : str, Path, optional
        Absolute path to the root containing the downloaded archive and meta.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Download link: https://zenodo.org/record/7262581/files/amos22.zip

    Examples
    --------
    >>> # Download the archive and meta to any folder and pass the path to the constructor:
    >>> ds = AMOS(root='/path/to/the/downloaded/files')
    >>> print(len(ds.ids))
    # 961
    >>> print(ds.image(ds.ids[0]).shape)
    # (768, 768, 90)
    >>> print(ds.mask(ds.ids[26]).shape)
    # (512, 512, 124)

    References
    ----------
    .. [1] JI YUANFENG. (2022). Amos: A large-scale abdominal multi-organ benchmark for
    versatile medical image segmentation [Data set]. Zenodo. https://doi.org/10.5281/zenodo.7262581
    """

    @property
    def ids(self):
        ids = list(self._id2split)

        for archive in [
            'amos22_unlabeled_ct_5000_5399.zip',
            'amos22_unlabeled_ct_5400_5899.zip',
            'amos22_unlabeled_ct_5900_6199.zip',
            'amos22_unlabeled_ct_6200_6899.zip',
        ]:
            file = self.root / archive
            if not file.exists():
                continue

            with ZipFile(file) as zf:
                for x in zf.namelist():
                    if x.endswith('.nii.gz'):
                        file = x.split('/')[-1]

                        ids.append(file.split('.')[0].split('_')[-1])

        return sorted(ids)

    @field
    def image(self, i):
        """Corresponding 3D image."""
        if i in ERRORS:
            return None  # this image is damaged in the archive

        archive_name, archive_root, file = self._archive_name(i)
        with unpack(self.root / archive_name, file, archive_root, '.zip') as (unpacked, is_unpacked):
            if is_unpacked:
                return np.asarray(nibabel.load(unpacked).dataobj)
            else:
                with open_nii_gz_file(unpacked) as image:
                    return np.asarray(image.dataobj)

    @field
    def affine(self, i):
        """The 4x4 matrix that gives the image's spatial orientation."""
        if i in ERRORS:
            return None  # this image is damaged in the archive

        archive_name, archive_root, file = self._archive_name(i)
        with unpack(self.root / archive_name, file, archive_root, '.zip') as (unpacked, is_unpacked):
            if is_unpacked:
                return nibabel.load(unpacked).affine
            else:
                with open_nii_gz_file(unpacked) as image:
                    return image.affine

    @field
    def mask(self, i):
        if i not in self._id2split:
            return

        file = f'labels{self._id2split[i]}/amos_{i}.nii.gz'
        try:
            with unpack(self.root / ARCHIVE_NAME_SEG, file, ARCHIVE_ROOT_NAME, '.zip') as (unpacked, is_unpacked):
                if is_unpacked:
                    return np.asarray(nibabel.load(unpacked).dataobj)
                else:
                    with open_nii_gz_file(unpacked) as image:
                        return np.asarray(image.dataobj)
        except FileNotFoundError:
            return

    @field
    def image_modality(self, i):
        """Returns image modality, `CT` or `MRI`."""
        if 500 < int(i) <= 600:
            return 'MRI'
        return 'CT'

    # labels
    @field
    def birth_date(self, i):
        return self._label(i, "Patient's Birth Date")

    @field
    def sex(self, i):
        return self._label(i, "Patient's Sex")

    @field
    def age(self, i):
        return self._label(i, "Patient's Age")

    @field
    def manufacturer_model(self, i):
        return self._label(i, "Manufacturer's Model Name")

    @field
    def manufacturer(self, i):
        return self._label(i, 'Manufacturer')

    @field
    def acquisition_date(self, i):
        return self._label(i, 'Acquisition Date')

    @field
    def site(self, i):
        return self._label(i, 'Site')

    @cached_property
    @composed(dict)
    def _id2split(self):
        with ZipFile(self.root / ARCHIVE_NAME_SEG) as zf:
            for x in zf.namelist():
                if (len(x.strip('/').split('/')) == 3) and x.endswith('.nii.gz'):
                    file, split = x.split('/')[-1], x.split('/')[-2][-2:]
                    id_ = file.split('.')[0].split('_')[-1]

                    yield id_, split

    @cached_property
    def _meta(self):
        files = [
            'labeled_data_meta_0000_0599.csv',
            'unlabeled_data_meta_5400_5899.csv',
            'unlabeled_data_meta_5000_5399.csv',
            'unlabeled_data_meta_5900_6199.csv',
        ]

        dfs = []
        for file in files:
            with unpack(self.root, file) as (unpacked, _):
                dfs.append(pd.read_csv(unpacked))
        return pd.concat(dfs)

    def _archive_name(self, i):
        if i in self._id2split:
            return ARCHIVE_NAME_SEG, ARCHIVE_ROOT_NAME, f'images{self._id2split[i]}/amos_{i}.nii.gz'

        i = int(i)
        file = f'amos_{i}.nii.gz'
        if 5000 <= i < 5400:
            return 'amos22_unlabeled_ct_5000_5399.zip', 'amos_unlabeled_ct_5000_5399', file
        elif 5400 <= i < 5900:
            return 'amos22_unlabeled_ct_5400_5899.zip', 'amos_unlabeled_ct_5400_5899', file
        elif 5900 <= i < 6200:
            return 'amos22_unlabeled_ct_5900_6199.zip', 'amos22_unlabeled_ct_5900_6199', file
        else:
            return 'amos22_unlabeled_ct_6200_6899.zip', 'amos22_unlabeled_6200_6899', file

    def _label(self, i, column):
        # ambiguous data in meta
        if int(i) in [500, 600]:
            return None
        elif int(i) not in self._meta['amos_id']:
            return None

        return self._meta[self._meta['amos_id'] == int(i)][column].item()



import numpy as np
from pylidc import Annotation

from .typing import (
    Calcification,
    InternalStructure,
    LIDCNodule,
    Lobulation,
    Malignancy,
    Sphericity,
    Spiculation,
    Subtlety,
    Texture,
)


def get_nodule(ann: Annotation) -> LIDCNodule:
    def init_enum(enum_class, value):
        try:
            return enum_class(value)
        except ValueError:
            pass

    bbox = ann.bbox_matrix().T
    bbox[1] = bbox[1] + 1

    return LIDCNodule(
        center_voxel=ann.centroid,
        bbox=bbox,
        diameter_mm=ann.diameter,
        surface_area_mm2=ann.surface_area,
        volume_mm3=ann.volume,
        calcification=init_enum(Calcification, ann.calcification),
        internal_structure=init_enum(InternalStructure, ann.internalStructure),
        lobulation=init_enum(Lobulation, ann.lobulation),
        malignancy=init_enum(Malignancy, ann.malignancy),
        sphericity=init_enum(Sphericity, ann.sphericity),
        spiculation=init_enum(Spiculation, ann.spiculation),
        subtlety=init_enum(Subtlety, ann.subtlety),
        texture=init_enum(Texture, ann.texture),
    )


def flip_nodule(nodule: LIDCNodule, n_slices: int) -> LIDCNodule:
    bbox = nodule.bbox.copy()
    start_slice, stop_slice = bbox[:, -1]
    bbox[:, -1] = np.array([n_slices - stop_slice, n_slices - start_slice])

    center_voxel = nodule.center_voxel
    center_voxel[-1] = n_slices - center_voxel[-1]

    return nodule._replace(
        center_voxel=center_voxel,
        bbox=bbox,
    )



import nibabel
import numpy as np

from ..internals.dataset import register_field
from ..utils import open_nii_gz_file, unpack
from .const import ANATOMICAL_STRUCTURES, LABELS


ARCHIVE_ROOT = 'Totalsegmentator_dataset'


def label_loader(name):
    def loader(self, i):
        return self._meta[self._meta['image_id'] == i][name].item()

    register_field('Totalsegmentator', name, loader)
    return loader


def mask_loader(name):
    def loader(self, i):
        file = f'{i}/segmentations/{name}.nii.gz'

        with unpack(self.root, file, ARCHIVE_ROOT, '.zip') as (unpacked, is_unpacked):
            if is_unpacked:
                return np.asarray(nibabel.load(unpacked).dataobj)
            else:
                with open_nii_gz_file(unpacked) as image:
                    return np.asarray(image.dataobj)

    register_field('Totalsegmentator', name, loader)
    return loader


def add_labels(scope):
    for label in LABELS:
        scope[label] = label_loader(label)


def add_masks(scope):
    for anatomical_structure in ANATOMICAL_STRUCTURES:
        scope[anatomical_structure] = mask_loader(anatomical_structure)



import gzip
from pathlib import Path

import nibabel as nb
import numpy as np

from .internals import Dataset, field, licenses, register


@register(
    body_region='Chest',
    license=licenses.CC_BYNC_40,
    link='https://github.com/XiaoweiXu/Dataset_Type-B-Aortic-Dissection',
    modality='CT',
    prep_data_size='14G',
    raw_data_size='14G',
    task='Aortic dissection segmentation',
)
class TBAD(Dataset):
    """
    A dataset of 3D Computed Tomography (CT) images for Type-B Aortic Dissection segmentation.

    Notes
    -----
    The data can only be obtained by contacting the authors by email.
    See the [dataset home page](https://github.com/XiaoweiXu/Dataset_Type-B-Aortic-Dissection) for details.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded files.
        If not provided, the cache is assumed to be already populated.

    Examples
    --------
    >>> # Place the downloaded files in any folder and pass the path to the constructor:
    >>> ds = TBAD(root='/path/to/files/root')
    >>> print(len(ds.ids))
    # 100
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 327)

    References
    ----------
    .. [1] Yao, Zeyang & Xie, Wen & Zhang, Jiawei & Dong, Yuhao & Qiu, Hailong & Haiyun, Yuan & Jia,
           Qianjun & Tianchen, Wang & Shi, Yiyi & Zhuang, Jian & Que, Lifeng & Xu, Xiaowei & Huang, Meiping.
           (2021). ImageTBAD: A 3D Computed Tomography Angiography Image Dataset for Automatic Segmentation
           of Type-B Aortic Dissection. Frontiers in Physiology. 12. 732711. 10.3389/fphys.2021.732711.
    """

    @property
    def ids(self):
        result = set()

        for file in self.root.glob('*_image.nii.gz'):
            result.add(file.stem.split('_')[0])

        return tuple(sorted(result))

    def _fname(self, i):
        return self.root / f'{i}_image.nii.gz'

    def image(self, i) -> np.ndarray:
        with self._fname(i).open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nb.FileHolder(fileobj=nii)
                image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return np.int16(image.get_fdata())

    def affine(self, i) -> np.ndarray:
        """The 4x4 matrix that gives the image's spatial orientation."""
        with self._fname(i).open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nb.FileHolder(fileobj=nii)
                image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return image.affine

    @field
    def mask(self, i) -> np.ndarray:
        with Path(self.root / f'{i}_label.nii.gz').open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nb.FileHolder(fileobj=nii)
                label = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return np.uint8(label.get_fdata())



import json

import numpy as np
import pandas as pd
import pydicom
from dicom_csv import (
    get_common_tag,
    get_pixel_spacing,
    get_slice_locations,
    get_tag,
    join_tree,
    order_series,
    stack_images,
)
from scipy import stats
from skimage.draw import polygon

from ..internals import Dataset, licenses, register
from ..utils import get_series_date


@register(
    body_region='Head',
    license=licenses.CC_BY_40,
    link='https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=70229053',
    modality=('MRI T1c', 'MRI T2'),
    prep_data_size='14,42G',
    raw_data_size='27G',
    task='Segmentation',
)
class VSSEG(Dataset):
    """
    Segmentation of vestibular schwannoma from MRI, an open annotated dataset ... (VS-SEG) [1]_.

    The dataset contains 250 pairs of T1c and T2 images of the brain with the
    vestibular schwannoma segmentation task.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.


    Notes
    -----
    The dataset and corresponding metadata could be downloaded at the TCIA page:
    https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=70229053.

    To download DICOM images using `.tcia` file, we used public build of TCIA downloader:
    https://github.com/ygidtu/NBIA_data_retriever_CLI.

    Then, download the rest of metadata from TCIA page:
      - `DirectoryNamesMappingModality.csv`
      - `Vestibular-Schwannoma-SEG_matrices Mar 2021.zip`
      - `Vestibular-Schwannoma-SEG contours Mar 2021.zip`

    and unzip the latter two `.zip` archives.

    So the `root` folder should contain 3 folders and 1 `.csv` file:
        <...>/DirectoryNamesMappingModality.csv
        <...>/Vestibular-Schwannoma-SEG/
                ├── VS-SEG-001/...
                ├── VS-SEG-002/...
                └── ...
        <...>/contours/
        <...>/registration_matrices/

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = VSSEG(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 484
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 120)
    >>> print(ds.schwannoma(ds.ids[1]).shape)
    # (384, 384, 80)

    References
    ----------
    .. [1] Shapey, Jonathan, et al. "Segmentation of vestibular schwannoma from MRI,
           an open annotated dataset and baseline algorithm."
           Scientific Data 8.1 (2021): 1-6.
           https://www.nature.com/articles/s41597-021-01064-w

    """

    @property
    def ids(self):
        subject_id_paths = list((self.root / 'Vestibular-Schwannoma-SEG').glob('VS-SEG-*'))
        t1_ids = [p.name + '-T1' for p in subject_id_paths]
        t2_ids = [p.name + '-T2' for p in subject_id_paths]
        return tuple(sorted(t1_ids + t2_ids))

    def modality(self, i):
        return i.rsplit('-', 1)[1]

    def subject_id(self, i):
        return i.rsplit('-', 1)[0]

    # ### series and images: ###

    def _series(self, i):
        return _load_series(i, self.root)

    def image(self, i):
        return stack_images(self._series(i), -1).transpose(1, 0, 2)

    # ### contours and masks ###

    def _contours(self, i):
        subject_num = int(self.subject_id(i).rsplit('-', 1)[-1])
        for file in (self.root / 'contours').glob(f'vs_gk_{subject_num}_{self.modality(i).lower()}/*'):
            with open(str(file), 'r') as f:
                return json.load(f)

    def _pixel_spacing(self, i):
        return get_pixel_spacing(self, i).tolist()

    def _slice_locations(self, i):
        return get_slice_locations(self, i)

    def spacing(self, i):
        """The maximum relative difference in `slice_locations` < 1e-12,
        so we allow ourselves to use the common spacing for the whole 3D image."""
        return (*self._pixel_spacing(i), stats.mode(np.diff(self._slice_locations(i)))[0].item())

    def _normed_contours(self, i):
        return _norm_contours(self._contours(i), self.spacing(i), self._patient_position(i))

    def _patient_position(self, i):
        return tuple(map(float, get_tag(self._series(i)[0], 'ImagePositionPatient')))

    def schwannoma(self, i):
        return _get_mask(self._normed_contours(i), self.image(i).shape, obj='schwannoma')

    def cochlea(self, i):
        return _get_mask(self._normed_contours(i), self.image(i).shape, obj='cochlea')

    def meningioma(self, i):
        return _get_mask(self._normed_contours(i), self.image(i).shape, obj='meningioma')

    # ### other DICOM metadata: ###

    def study_uid(self, i):
        return get_common_tag(self._series(i), 'StudyInstanceUID')

    def series_uid(self, i):
        return get_common_tag(self._series(i), 'SeriesInstanceUID')

    def patient_id(self, i):
        return get_common_tag(self._series(i), 'PatientID', default=None)

    def study_date(self, i):
        return get_series_date(self, i)


def _load_series(_id, root):
    _id, modality = _id.rsplit('-', 1)
    df = pd.read_csv(root / 'DirectoryNamesMappingModality.csv')
    df = df[df['Classic Directory Name'].apply(lambda x: _id in x)]

    study_id, series_id = df[df['Modality'] == f'{modality} image']['Classic Directory Name'].iloc[0].split('/')[1:]
    next_single_folder = list((root / 'Vestibular-Schwannoma-SEG' / _id / study_id).glob('*'))[0].name
    path_to_series = root / 'Vestibular-Schwannoma-SEG' / _id / study_id / next_single_folder / series_id
    tree = join_tree(path_to_series)
    tree = tree[tree['NoError']]

    series = [pydicom.dcmread(path_to_series / fname) for fname in tree['FileName'].values]
    series = order_series(series, decreasing=False)
    return series


def _norm_contours(contours, voxel_spacing, origin):
    voxel_spacing = np.float32(voxel_spacing)
    pixel_spacing, z_spacing = voxel_spacing[:-1], voxel_spacing[-1]
    origin = np.float32(origin)

    contours_normed = []

    for contour in contours:
        cs = contour['LPS_contour_points']
        cs_normed = []
        for c in cs:
            c = np.float32(c)
            xy_normed = (c[:, :-1] - origin[:-1]) / pixel_spacing
            z_normed = np.round((c[0][-1] - origin[-1]) / z_spacing)
            c_normed = np.concatenate((xy_normed, np.repeat([[z_normed]], xy_normed.shape[0], axis=0)), axis=-1)
            cs_normed.append(c_normed.tolist())
        contours_normed.append({**contour, 'LPS_contour_points': cs_normed})

    return contours_normed


def _contour2mask(cnt: list, shape):
    msk = np.zeros(shape, dtype=bool)
    for planar_c in cnt:
        idx = int(planar_c[0][-1])
        slc = msk[..., idx]
        r, c, _ = np.transpose(planar_c)
        rr, cc = polygon(r, c, shape=slc.shape)
        slc[rr, cc] = 1
    return msk


def _contours2names(contours: list):
    return [c['structure_name'] for c in contours]


def _select_contour_by_structure_name(contours: list, name: str):
    return [c for c in contours if c['structure_name'] == name][0]['LPS_contour_points']


def _get_schwannoma_structure_name(contours: list):
    names = _contours2names(contours)

    # filter 1:
    filter_names = (
        'Brainstem',
        'Modiolus',  # CS (other)
        'Cochlea',
        'Cochlea_c',
        'Cochlea_d',
        'cochlea',  # CS (Cochlea)
        'Test',  # a duplicate of 'TV' with Dice Score = 0.92 (seems to be a worse contour)
        'Men',  # is a meningioma case
    )
    names = [name for name in names if name not in filter_names]

    if len(names) == 0:
        return None

    if len(names) == 1:
        if names[0] == 'vol 2y':
            # 'VS-SEG-063' -- is a meningioma case + "vol 2y" contour has a bad interpolation:(
            return None
        else:
            return names[0]

    # filter 2:
    # no 'TV' simultaneously with 'AN'
    higher_priority_names = (
        'TV',
        'tv',
        'AN',
    )
    is_high_priority = [name in higher_priority_names for name in names]
    if np.sum(is_high_priority) == 1:
        return np.array(names)[np.array(is_high_priority)][0]

    return names


def _get_cochlea_structure_name(contours: list):
    names = _contours2names(contours)
    cochlea_names = ('Cochlea', 'Cochlea_c', 'Cochlea_d', 'cochlea')
    # there are no duplicated cochlea structures:
    for name in names:
        if name in cochlea_names:
            return name


def _get_meningioma_structure_name(contours: list):
    return 'Men' if 'Men' in _contours2names(contours) else None


def _get_mask(contours: list, shape, obj: str):
    get_structure_name = {
        'schwannoma': _get_schwannoma_structure_name,
        'cochlea': _get_cochlea_structure_name,
        'meningioma': _get_meningioma_structure_name,
    }[obj]
    structure_name = get_structure_name(contours)
    if structure_name is not None:
        contour = _select_contour_by_structure_name(contours, structure_name)
        return _contour2mask(contour, shape)



import importlib
import inspect
from collections import OrderedDict
from pathlib import Path
from typing import NamedTuple, Type

import pandas as pd

from .licenses import License


_REGISTRY = {}


class Description(NamedTuple):
    body_region: str = None
    license: str = None
    link: str = None
    modality: str = None
    prep_data_size: str = None
    raw_data_size: str = None
    task: str = None


def register(**kwargs):
    def decorator(cls: Type):
        _register(cls, cls.__name__, description, 2)
        # cls._path = path
        return cls

    # path = kwargs.pop('path')
    description = Description(**kwargs)
    return decorator


def _register(cls, name, description, level):
    module = inspect.getmodule(inspect.stack()[level][0]).__name__
    assert name not in _REGISTRY, name
    _REGISTRY[name] = cls, module, description


def gather_datasets():
    for f in Path(__file__).resolve().parent.parent.iterdir():
        module_name = f'amid.{f.stem}'
        importlib.import_module(module_name)

    return OrderedDict((k, _REGISTRY[k]) for k in sorted(_REGISTRY))


def prepare_for_table(name, count, module, description, version):
    def stringify(x):
        if pd.isnull(x):
            return ''
        if isinstance(x, str):
            return x
        if isinstance(x, (list, tuple)):
            return ', '.join(x)
        return x

    entry = {'name': name, 'entries': count}
    entry.update({k: v for k, v in description._asdict().items() if not pd.isnull(v)})
    license_ = entry.get('license', None)
    if license_:
        if isinstance(license_, License):
            license_ = f'<a href="{license_.url}">{license_.name}</a>'
        entry['license'] = license_

    link = entry.pop('link', None)
    if link is not None:
        entry['link'] = f'<a href="{link}">Source</a>'

    entry['name'] = f'<a href="https://neuro-ml.github.io/amid/{version}/datasets-api/#{module}.{name}">{name}</a>'
    return {k: stringify(v) for k, v in entry.items()}



import contextlib
import gzip
import re
import zipfile
from pathlib import Path
from zipfile import ZipFile

import nibabel as nb
import numpy as np

from .internals import Dataset, field, licenses, register


@register(
    body_region=('Chest', 'Abdomen'),
    license=licenses.CC_BYSA_40,
    link='https://www.medseg.ai/database/liver-segments-50-cases',
    modality='CT',
    prep_data_size='1,88G',
    raw_data_size='616M',
    task='Segmentation',
)
class LiverMedseg(Dataset):
    """
    LiverMedseg is a public CT segmentation dataset with 50 annotated images.
    Case collection of 50 livers with their segments.
    Images obtained from Decathlon Medical Segmentation competition

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Download links:
    https://www.medseg.ai/database/liver-segments-50-cases

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = LiverMedseg(root='/path/to/archives/root')
    >>> print(len(ds.ids))
    # 50
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 38)

    References
    ----------
    """

    @property
    def ids(self):
        result = set()
        with ZipFile(self.root / 'img.zip') as zf:
            for zipinfo in zf.infolist():
                if zipinfo.is_dir():
                    continue
                file_stem = Path(zipinfo.filename).stem
                result.add('liver_medseg_' + re.findall(r'\d+', file_stem)[0])

        return tuple(sorted(result))

    def _file(self, i):
        num_id = i.split('_')[-1]
        return zipfile.Path(self.root / 'img.zip', f'img{num_id}.nii.gz')

    @field
    def image(self, i) -> np.ndarray:
        with open_nii_gz_file(self._file(i)) as nii_file:
            return np.asarray(nii_file.dataobj)

    @field
    def affine(self, i) -> np.ndarray:
        """The 4x4 matrix that gives the image's spatial orientation."""
        with open_nii_gz_file(self._file(i)) as nii_file:
            return nii_file.affine

    def spacing(self, i) -> tuple:
        with open_nii_gz_file(self._file(i)) as nii_file:
            return tuple(nii_file.header['pixdim'][1:4])

    @field
    def mask(self, i) -> np.ndarray:
        path = Path(str(self._file(i)).replace('img', 'mask'))
        folder, image = path.parent, path.name
        _file = zipfile.Path(folder, image)
        with open_nii_gz_file(_file) as nii_file:
            return np.asarray(nii_file.dataobj).astype(np.uint8)


# TODO: sync with amid.utils
@contextlib.contextmanager
def open_nii_gz_file(file):
    with file.open('rb') as opened:
        with gzip.GzipFile(fileobj=opened) as nii:
            nii = nb.FileHolder(fileobj=nii)
            yield nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})



import json
import os.path
import warnings
from functools import cached_property
from typing import Tuple

import numpy as np
import pandas as pd
import pydicom
from dicom_csv import (
    drop_duplicated_instances,
    drop_duplicated_slices,
    expand_volumetric,
    get_pixel_spacing,
    get_slice_locations,
    join_tree,
    order_series,
    stack_images,
)
from skimage.draw import polygon

from .internals import Dataset, field, licenses, register


@register(
    body_region='Thorax',
    license=licenses.CC_BYNC_40,
    link='https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=80969742',
    modality='CT',
    prep_data_size='7,9G',
    raw_data_size='12G',
    task='COVID-19 Segmentation',
)
class MIDRC(Dataset):
    """

        MIDRC-RICORD dataset 1a is a public COVID-19 CT segmentation dataset with 120 scans.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Follow the download instructions at https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=80969742
    Download both Images and Annotations to the same folder

    Then, the folder with downloaded data should contain two paths with the data

    The folder should have this structure:
        <...>/<MIDRC-root>/MIDRC-RICORD-1A
        <...>/<MIDRC-root>/MIDRC-RICORD-1a_annotations_labelgroup_all_2020-Dec-8.json


    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = MIDRC(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
     155
    >>> print(ds.image(ds.ids[0]).shape)
     (512, 512, 112)
    >>> print(ds.mask(ds.ids[80]).shape)
     (6, 512, 512, 450)

    References
    ----------
    """

    _fields = 'image', 'image_meta', 'spacing', 'labels', 'mask'

    @cached_property
    def _joined(self):
        joined_path = self.root / 'joined.csv'
        if joined_path.exists():
            joined = pd.read_csv(joined_path)
        else:
            joined = join_tree(self.root / 'MIDRC-RICORD-1A', verbose=1)
            joined = joined[[x.endswith('.dcm') for x in joined.FileName]]
            joined.to_csv(self.root / 'joined.csv', index=False)
        return joined

    @cached_property
    def ids(self):
        return tuple(self._joined['SeriesInstanceUID'].unique())

    @cached_property
    def _annotation(self):
        json_path = 'MIDRC-RICORD-1a_annotations_labelgroup_all_2020-Dec-8.json'
        return json_to_dataframe(self.root / json_path)['annotations']

    def _series(self, i):
        sub = self._joined[self._joined.SeriesInstanceUID == i]
        series_files = sub['PathToFolder'] + os.path.sep + sub['FileName']
        series_files = [self.root / 'MIDRC-RICORD-1A' / x for x in series_files]
        series = list(map(pydicom.dcmread, series_files))
        # series = sorted(series, key=lambda x: x.InstanceNumber)
        series = expand_volumetric(series)
        series = drop_duplicated_instances(series)

        # if drop_dupl_slices:
        _original_num_slices = len(series)
        series = drop_duplicated_slices(series)
        if len(series) < _original_num_slices:
            warnings.warn(f'Dropped duplicated slices for series {series[0]["StudyInstanceUID"]}.')

        series = order_series(series, decreasing=False)
        return series

    @field
    def image(self, i):
        image = stack_images(self._series(i), -1).astype(np.int16).transpose(1, 0, 2)
        return image

    @field
    def image_meta(self, i):
        metas = [list(dict(s).values()) for s in self._series(i)]
        result = {}
        for meta_ in metas:
            for element in meta_:
                if element.keyword in ['PixelData']:
                    continue
                if element.keyword not in result:
                    result[element.keyword] = [element.value]
                elif result[element.keyword][-1] != element.value:
                    result[element.keyword].append(element.value)
        # turn elements that are the same across the series back from array
        result = {k: v[0] if len(v) == 1 else v for k, v in result.items()}
        return result

    def _study_id(self, i):
        study_ids = self._joined[self._joined.SeriesInstanceUID == i].StudyInstanceUID.unique()
        assert len(study_ids) == 1
        # series_id_to_study
        return study_ids[0]

    @field
    def spacing(self, i):
        series = self._series(i)
        pixel_spacing = get_pixel_spacing(series).tolist()
        slice_locations = get_slice_locations(series)
        diffs, counts = np.unique(np.round(np.diff(slice_locations), decimals=5), return_counts=True)
        spacing = np.float32([pixel_spacing[1], pixel_spacing[0], diffs[np.argsort(counts)[-1]]])
        return tuple(spacing.tolist())

    @field
    def labels(self, i):
        sub = self._annotation[
            (self._annotation.scope == 'STUDY') & (self._annotation.StudyInstanceUID == self._study_id(i))
        ]
        return tuple(sub['labelName'].unique())

    @field
    def mask(self, i):
        # TODO: mask has 6 channels now. Consider adding different methods ot at least a docstring naming channels...
        sub = self._annotation[(self._annotation.SeriesInstanceUID == i) & (self._annotation.scope == 'INSTANCE')]
        image_meta = self.image_meta(i)

        shape = (image_meta['Rows'], image_meta['Columns'], len(image_meta['SOPInstanceUID']))
        mask = np.zeros((len(_pathologies), *shape), dtype=bool)
        if len(sub) == 0:
            return None
        for label, row in sub.iterrows():
            pathology_index = _pathologies.index(row['labelName'])
            slice_index = image_meta['SOPInstanceUID'].index(row['SOPInstanceUID'])
            if row['data'] is None:
                warnings.warn(f'{label} annotations for series {i} contains None for slice {slice_index}.')
                continue
            ys, xs = np.array(row['data']['vertices']).T
            mask[(pathology_index, *polygon(ys, xs, shape[:2]), slice_index)] = True
        return mask


_pathologies: Tuple[str, ...] = (
    'Infectious opacity',
    'Infectious TIB/micronodules',
    'Atelectasis',
    'Other noninfectious opacity',
    'Noninfectious nodule/mass',
    'Infectious cavity',
)


# TODO: simplify
def json_to_dataframe(json_file, datasets=None):
    if datasets is None:
        datasets = []
    with open(json_file, 'r', encoding='utf-8') as f:
        data = json.load(f)

    a = pd.DataFrame([])
    studies = pd.DataFrame([])
    labels = None

    # Gets annotations for all datasets
    for d in data['datasets']:
        if d['id'] in datasets or len(datasets) == 0:
            study = pd.DataFrame(d['studies'])
            study['dataset'] = d['name']
            study['datasetId'] = d['id']
            studies = pd.concat([studies, study], ignore_index=True, sort=False)

            annots = pd.DataFrame(d['annotations'])
            annots['dataset'] = d['name']
            a = pd.concat([a, annots], ignore_index=True, sort=False)

    if len(studies) > 0:
        studies = studies[['StudyInstanceUID', 'dataset', 'datasetId', 'number']]
    g = pd.DataFrame(data['labelGroups'])
    # unpack arrays
    result = pd.DataFrame([(d, tup.id, tup.name) for tup in g.itertuples() for d in tup.labels])
    if len(result) > 0:
        result.columns = ['labels', 'labelGroupId', 'labelGroupName']

        def unpack_dictionary(df, column):
            ret = pd.concat([df, pd.DataFrame((d for idx, d in df[column].items()))], axis=1, sort=False)
            del ret[column]
            return ret

        labels = unpack_dictionary(result, 'labels')
        if 'parentId' in labels.columns:
            labels = labels[
                [
                    'labelGroupId',
                    'labelGroupName',
                    'annotationMode',
                    'color',
                    'description',
                    'id',
                    'name',
                    'radlexTagIds',
                    'scope',
                    'parentId',
                ]
            ]
            labels.columns = [
                'labelGroupId',
                'labelGroupName',
                'annotationMode',
                'color',
                'description',
                'labelId',
                'labelName',
                'radlexTagIdsLabel',
                'scope',
                'parentLabelId',
            ]
        else:
            labels = labels[
                [
                    'labelGroupId',
                    'labelGroupName',
                    'annotationMode',
                    'color',
                    'description',
                    'id',
                    'name',
                    'radlexTagIds',
                    'scope',
                ]
            ]
            labels.columns = [
                'labelGroupId',
                'labelGroupName',
                'annotationMode',
                'color',
                'description',
                'labelId',
                'labelName',
                'radlexTagIdsLabel',
                'scope',
            ]

        if len(a) > 0:
            a = a.merge(labels, on=['labelId'], sort=False)
    if len(studies) > 0 and len(a) > 0:
        a = a.merge(studies, on=['StudyInstanceUID', 'dataset'], sort=False)
        # Format data
        studies.number = studies.number.astype(int)
        a.number = a.number.astype(int)
        a.loc.createdAt = pd.to_datetime(a.createdAt)
        a.loc.updatedAt = pd.to_datetime(a.updatedAt)
    return {'annotations': a, 'studies': studies, 'labels': labels}



from contextlib import suppress
from functools import cached_property

import pandas as pd
import pydicom

from ..internals import Dataset, field, register
from .utils import csv_field, unpack


@register(
    body_region='Thorax',
    license='Non-Commercial Use',
    link='https://www.kaggle.com/competitions/rsna-breast-cancer-detection/data',
    modality='MG',
    raw_data_size='271G',
    prep_data_size='294G',
    task='Breast cancer classification',
)
class RSNABreastCancer(Dataset):
    @cached_property
    def _meta(self):
        dfs = []
        for part in 'train', 'test':
            with suppress(FileNotFoundError):
                with unpack(self.root, f'{part}.csv') as (file, _):
                    df = pd.read_csv(file)
                    df['part'] = part
                    dfs.append(df)

        if not dfs:
            raise FileNotFoundError('No metadata found')
        dfs = pd.concat(dfs, ignore_index=True)
        for name in 'image_id', 'patient_id', 'site_id':
            dfs[name] = dfs[name].astype(str)

        raw = list(map(str, dfs.image_id.tolist()))
        ids = set(raw)
        if len(ids) != len(raw):
            raise ValueError('The image ids are not unique')

        return {row.image_id: row for _, row in dfs.iterrows()}

    # csv fields
    site_id = csv_field('site_id', str)
    patient_id = csv_field('patient_id', str)
    image_id = csv_field('image_id', str)
    laterality = csv_field('laterality', None)
    view = csv_field('view', None)
    age = csv_field('age', None)
    cancer = csv_field('cancer', bool)
    biopsy = csv_field('biopsy', bool)
    invasive = csv_field('invasive', bool)
    BIRADS = csv_field('BIRADS', int)
    implant = csv_field('implant', bool)
    density = csv_field('density', None)
    machine_id = csv_field('machine_id', str)
    prediction_id = csv_field('prediction_id', str)
    difficult_negative_case = csv_field('difficult_negative_case', bool)

    @property
    def ids(self):
        return tuple(sorted(self._meta))

    def _dicom(self, i):
        row = self._meta[i]
        with unpack(self.root, f'{row.part}_images/{row.patient_id}/{row.image_id}.dcm') as (file, _):
            return pydicom.dcmread(file)

    @field
    def image(self, i):
        return self._dicom(i).pixel_array

    @field
    def padding_value(self, i):
        return getattr(self._dicom(i), 'PixelPaddingValue', None)

    @field
    def intensity_sign(self, i):
        return getattr(self._dicom(i), 'PixelIntensityRelationshipSign', None)



import nibabel as nb
import numpy as np

from .internals import Dataset, field, register
from .utils import PathOrStr


@register(
    body_region='thorax',
    license=None,  # todo
    link='https://kits-challenge.org/kits23/',
    modality='CT',
    prep_data_size='50G',
    raw_data_size='12G',
    task='Kidney Tumor Segmentation',
)
class KiTS23(Dataset):
    """Kidney and Kidney Tumor Segmentation Challenge,
    The 2023 Kidney and Kidney Tumor Segmentation challenge (abbreviated KiTS23)
    is a competition in which teams compete to develop the best system for
    automatic semantic segmentation of kidneys, renal tumors, and renal cysts.

    Competition page is https://kits-challenge.org/kits23/, official competition repository is
    https://github.com/neheller/kits23/.

    For usage, clone the repository https://github.com/neheller/kits23/, install and run `kits23_download_data`.

    Parameters
    ----------
    root: str, Path
        Absolute path to the root containing the downloaded archive and meta.
        If not provided, the cache is assumed to be already populated.
    """

    def __init__(self, root: PathOrStr):
        super().__init__(root)
        if not (self.root / "dataset").exists():
            raise FileNotFoundError(f"Dataset not found in {self.root}")

    @property
    def ids(self):
        return tuple(sorted(sub.name for sub in (self.root / 'dataset').glob('*')))

    @field
    def image(self, i):
        # CT images are integer-valued, this will help us improve compression rates
        image_file = nb.load(self.root / 'dataset' / i / 'imaging.nii.gz')
        return np.int16(image_file.get_fdata()[...])

    # TODO add multiple segmentations
    @field
    def mask(self, i):
        """Combined annotation for kidneys, tumor and cyst (if present)."""
        ct_scan_nifti = nb.load(self.root / 'dataset' / i / 'segmentation.nii.gz')
        return np.int8(ct_scan_nifti.get_fdata())

    @field
    def affine(self, i):
        """The 4x4 matrix that gives the image's spatial orientation."""
        image_file = nb.load(self.root / 'dataset' / i / 'imaging.nii.gz')
        return image_file.affine

    @property
    def labels_names(self):
        """Indicates which label correspond to which mask, consistent accross all samples."""
        return KITS_LABEL_NAMES


KITS_LABEL_NAMES = {
    # https://github.com/neheller/kits23/blob/063d4c00afd383fc68145a00c0aa6a4e2a3c0f50/kits23/configuration/labels.py#L23
    1: 'kidney',
    2: 'tumor',
    3: 'cyst',
}



import contextlib
import gzip
import zipfile
from pathlib import Path
from zipfile import ZipFile

import nibabel as nb
import numpy as np

from ..internals import Dataset, licenses, register


@register(
    body_region='Head',
    license=licenses.CC_BYND_40,
    link='https://sites.google.com/view/calgary-campinas-dataset/home',
    modality='MRI T1',
    prep_data_size='14,66G',
    raw_data_size='4,1G',
    task='Segmentation',
)
class CC359(Dataset):
    """
    A (C)algary-(C)ampinas public brain MR dataset with (359) volumetric images [1]_.

    There are three segmentation tasks on this dataset: (i) brain, (ii) hippocampus, and
    (iii) White-Matter (WM), Gray-Matter (WM), and Cerebrospinal Fluid (CSF) segmentation.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.


    Notes
    -----
    homepage (upd): https://sites.google.com/view/calgary-campinas-dataset/home
    homepage (old): https://miclab.fee.unicamp.br/calgary-campinas-359-updated-05092017

    To obtain MR images and brain and hippocampus segmentation masks, please, follow the instructions
    at the download platform: https://portal.conp.ca/dataset?id=projects/calgary-campinas.

    Via `datalad` lib you need to download three zip archives:
        - `Original.zip` (the original MR images)
        - `hippocampus_staple.zip` (Silver-standard hippocampus masks generated using STAPLE)
        - `Silver-standard-machine-learning.zip` (Silver-standard brain masks generated using a machine learning method)

    To the current date, WM, GM, and CSF mask could be downloaded only from the google drive:
    https://drive.google.com/drive/u/0/folders/0BxLb0NB2MjVZNm9JY1pWNFp6WTA?resourcekey=0-2sXMr8q-n2Nn6iY3PbBAdA.

    Here you need to manually download a folder (from the google drive root above)
    `CC359/Reconstructed/CC359/WM-GM-CSF/`

    So the root folder to pass to this dataset class should contain four objects:
        - three zip archives (`Original.zip`, `hippocampus_staple.zip`, and `Silver-standard-machine-learning.zip`)
        - one folder `WM-GM-CSF` with the original structure:
            <...>/WM-GM-CSF/CC0319_ge_3_45_M.nii.gz
            <...>/WM-GM-CSF/CC0324_ge_3_56_M.nii.gz
            ...

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> cc359 = CC359(root='/path/to/downloaded/data/folder/')
    >>> print(len(cc359.ids))
    # 359
    >>> print(cc359.image(cc359.ids[0]).shape)
    # (171, 256, 256)
    >>> print(cc359.wm_gm_csf(cc359.ids[80]).shape)
    # (180, 240, 240)

    References
    ----------
    .. [1] Souza, Roberto, et al. "An open, multi-vendor, multi-field-strength brain MR dataset
           and analysis of publicly available skull stripping methods agreement."
           NeuroImage 170 (2018): 482-494.
           https://www.sciencedirect.com/science/article/pii/S1053811917306687

    """

    @property
    def ids(self):
        result = set()
        with ZipFile(self.root / 'Original.zip') as zf:
            for zipinfo in zf.infolist():
                if zipinfo.is_dir():
                    continue

                file_name = Path(zipinfo.filename).name
                if file_name.startswith('CC'):
                    result.add(file_name.split('_')[0])

        return tuple(sorted(result))

    def _image_file(self, i):
        return get_zipfile(i, 'Original.zip', self.root)

    def vendor(self, i):
        return zipfile2meta(self._image_file(i))['vendor']

    def field(self, i):
        return zipfile2meta(self._image_file(i))['field']

    def age(self, i):
        return zipfile2meta(self._image_file(i))['age']

    def sex(self, i):
        return zipfile2meta(self._image_file(i))['gender']

    def image(self, i):
        with open_nii_gz_file(self._image_file(i)) as nii_image:
            return np.asarray(nii_image.dataobj)

    def affine(self, i):
        """The 4x4 matrix that gives the image's spatial orientation."""
        with open_nii_gz_file(self._image_file(i)) as nii_image:
            return nii_image.affine

    def spacing(self, i):
        """Returns voxel spacing along axes (x, y, z)."""
        with open_nii_gz_file(self._image_file(i)) as nii_image:
            return tuple(nii_image.header['pixdim'][1:4])

    # masks:

    def brain(self, i):
        zf = get_zipfile(i, 'Silver-standard-machine-learning.zip', self.root)
        with open_nii_gz_file(zf) as nii_image:
            return np.uint8(nii_image.get_fdata())

    def hippocampus(self, i):
        try:
            zf = get_zipfile(i, 'hippocampus_staple.zip', self.root)
        except KeyError:
            return None

        with open_nii_gz_file(zf) as nii_image:
            return np.uint8(nii_image.get_fdata())

    def wm_gm_csf(self, i):
        for file in (self.root / 'WM-GM-CSF').glob('*'):
            if file.name.startswith(i):
                with open_nii_gz_file(file) as nii_image:
                    return np.uint8(nii_image.get_fdata())


# TODO: sync with amid.utils
@contextlib.contextmanager
def open_nii_gz_file(file):
    with file.open('rb') as opened:
        with gzip.GzipFile(fileobj=opened) as nii:
            nii = nb.FileHolder(fileobj=nii)
            yield nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})


def get_zipfile(_id, archive_name, root):
    archive = Path(root) / archive_name
    with ZipFile(archive) as zf:
        for zipinfo in zf.infolist():
            if Path(zipinfo.filename).name.startswith(_id) and not zipinfo.is_dir():
                return zipfile.Path(str(archive), zipinfo.filename)

    raise KeyError(f'Id "{_id}" not found')


def zipfile2meta(zf):
    return dict(zip(['id', 'vendor', 'field', 'age', 'gender'], zf.name[: -len('.nii.gz')].split('_')))



import zipfile
from pathlib import Path
from zipfile import ZipFile

import nibabel as nb
import numpy as np

from ..internals import Dataset, licenses, register


@register(
    body_region='Abdominal',
    license=licenses.CC_BYNCND_40,
    link='https://competitions.codalab.org/competitions/17094',
    modality='CT',
    prep_data_size='24,7G',
    raw_data_size='35G',
    task='Segmentation',
)
class LiTS(Dataset):
    """
    A (Li)ver (T)umor (S)egmentation dataset [1]_ from Medical Segmentation Decathlon [2]_

    There are two segmentation tasks on this dataset: liver and liver tumor segmentation.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.


    Notes
    -----
    Follow the download instructions at https://competitions.codalab.org/competitions/17094.

    Then, the folder with raw downloaded data should contain two zip archives with the train data
    (`Training_Batch1.zip` and `Training_Batch2.zip`)
    and a folder with the test data
    (`LITS-Challenge-Test-Data`).

    The folder with test data should have original structure:
        <...>/LITS-Challenge-Test-Data/test-volume-0.nii
        <...>/LITS-Challenge-Test-Data/test-volume-1.nii
        ...

    P.S. Organs boxes are also provided from a separate source https://github.com/superxuang/caffe_3d_faster_rcnn.

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = LiTS(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 201
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 163)
    >>> print(ds.tumor_mask(ds.ids[80]).shape)
    # (512, 512, 771)

    References
    ----------
    .. [1] Bilic, Patrick, et al. "The liver tumor segmentation benchmark (lits)."
           arXiv preprint arXiv:1901.04056 (2019).
    .. [2] Antonelli, Michela, et al. "The medical segmentation decathlon."
           arXiv preprint arXiv:2106.05735 (2021).
    """

    @property
    def ids(self):
        result = set()
        # zip archives for train images:
        for archive in self.root.glob('*.zip'):
            with ZipFile(archive) as zf:
                for zipinfo in zf.infolist():
                    if zipinfo.is_dir():
                        continue

                    file_stem = Path(zipinfo.filename).stem
                    if 'volume' in file_stem:
                        result.add('lits-train-' + file_stem.split('-')[-1])

        # folder for test images:
        for file in (self.root / 'LITS-Challenge-Test-Data').glob('*'):
            result.add('lits-test-' + file.stem.split('-')[-1])

        return tuple(sorted(result))

    def fold(self, i):
        num_id = i.split('-')[-1]

        if 'train' in i:
            for archive in self.root.glob('*.zip'):
                batch = '1' if ('1' in archive.stem) else '2'

                with ZipFile(archive) as zf:
                    for zipinfo in zf.infolist():
                        if zipinfo.is_dir():
                            continue

                        if num_id == Path(zipinfo.filename).stem.split('-')[-1]:
                            return f'train_batch_{batch}'

        else:  # if 'test' in i:
            return 'test'

    def _file(self, i):
        num_id = i.split('-')[-1]

        if 'train' in i:
            for archive in self.root.glob('*.zip'):
                with ZipFile(archive) as zf:
                    for zipinfo in zf.infolist():
                        if zipinfo.is_dir():
                            continue

                        file = Path(zipinfo.filename)
                        if ('volume' in file.stem) and (num_id == file.stem.split('-')[-1]):
                            return zipfile.Path(str(archive), str(file))

        else:  # if 'test' in i:
            return self.root / 'LITS-Challenge-Test-Data' / f'test-volume-{num_id}.nii'

        raise KeyError(f'Id "{i}" not found')

    def image(self, i):
        with self._file(i).open('rb') as nii:
            nii = nb.FileHolder(fileobj=nii)
            image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
            # most ct scans are integer-valued, this will help us improve compression rates
            return np.int16(image.get_fdata())

    def affine(self, i):
        """The 4x4 matrix that gives the image's spatial orientation."""
        with self._file(i).open('rb') as nii:
            nii = nb.FileHolder(fileobj=nii)
            image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
            return image.affine

    def spacing(self, i):
        """Returns voxel spacing along axes (x, y, z)."""
        with self._file(i).open('rb') as nii:
            nii = nb.FileHolder(fileobj=nii)
            image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
            return tuple(image.header['pixdim'][1:4])

    def mask(self, i):
        file = self._file(i)
        if 'test' not in file.name:
            with (file.parent / file.name.replace('volume', 'segmentation')).open('rb') as nii:
                nii = nb.FileHolder(fileobj=nii)
                image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return np.uint8(image.get_fdata())



from typing import NamedTuple


class ClinicalInfo(NamedTuple):
    gender: str
    age_at_scan_years: float
    survival_from_surgery_days: int
    idh1: str
    mgmt: str
    kps: str
    gtr_over90percent: str
    time_since_baseline_preop: int
    psp_tp_score: float


class AcquisitionInfo(NamedTuple):
    manufacturer: str
    model: str
    magnetic_field_strength: float
    t1_imaging_frequency: float
    t1_repetition_time: float
    t1_echo_time: float
    t1_inversion_time: float
    t1_flip_angle: float
    t1_pixel_spacing: str
    t1_slice_thickness: float
    t1gd_imaging_frequency: float
    t1gd_repetition_time: float
    t1gd_echo_time: float
    t1gd_inversion_time: float
    t1gd_flip_angle: float
    t1gd_pixel_spacing: str
    t1gd_slice_thickness: float
    t2_imaging_frequency: float
    t2_repetition_time: float
    t2_echo_time: float
    t2_flip_angle: float
    t2_pixel_spacing: str
    t2_slice_thickness: float
    flair_imaging_frequency: float
    flair_repetition_time: float
    flair_echo_time: float
    flair_inversion_time: float
    flair_flip_angle: float
    flair_pixel_spacing: str
    flair_slice_thickness: float
    dti_imaging_frequency: float
    dti_repetition_time: float
    dti_echo_time: float
    dti_flip_angle: float
    dti_pixel_spacing: str
    dti_slice_thickness: float
    dsc_imaging_frequency: float
    dsc_repetition_time: float
    dsc_echo_time: float
    dsc_flip_angle: float
    dsc_pixel_spacing: str
    dsc_slice_thickness: float



from pathlib import Path
from typing import Sequence

from connectome import ExternalBase

from ..utils import PathOrStr


class Dataset(ExternalBase):
    _path: str
    _fields: Sequence[str] = None

    def __init__(self, root: PathOrStr):
        fields = None
        if hasattr(self, '_fields'):
            fields = self._fields

        super().__init__(fields=fields, inputs=['id'], properties=['ids'], inherit=['id'])
        self.root = Path(root)

    @classmethod
    def __getversion__(cls):
        return 0


_Fields = {}


def register_field(cls, name, func):
    _Fields.setdefault(cls, {})[name] = func


def field(func):
    cls, name = func.__qualname__.split('.')
    register_field(cls, name, func)
    return func



from enum import Enum
from typing import NamedTuple, Optional, Sequence


class Texture(Enum):
    Solid, PartSolid, GroundGlass, Other = 0, 1, 2, 3


class Review(Enum):
    Confirmed, ConfirmedPartially, Doubt, Rejected = 0, 1, 2, 3


class Comment(Enum):
    Fibrosis, LymphNode, Calcium, Calcified, Bronchiectasis, Vessel = 0, 1, 2, 3, 4, 5


class Cancer500Nodule(NamedTuple):
    center_voxel: Sequence[int]
    review: Review
    comment: Optional[Comment] = None
    diameter_mm: Optional[float] = None
    texture: Optional[Texture] = None
    malignancy: Optional[bool] = None



import gzip
import zipfile
from pathlib import Path
from typing import Union
from zipfile import ZipFile

import nibabel
import numpy as np

from .internals import Dataset, field, register


@register(
    body_region='Abdomen',
    license=None,
    link='https://flare22.grand-challenge.org/',
    modality='CT',
    prep_data_size='347G',
    raw_data_size='247G',
    task='Semi-supervised abdominal organ segmentation',
)
class FLARE2022(Dataset):
    """
    An abdominal organ segmentation dataset for semi-supervised learning [1]_.

    The dataset was used at the MICCAI FLARE 2022 challenge.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Download link: https://flare22.grand-challenge.org/Dataset/

    The `root` folder should contain the two downloaded folders, namely: "Training" and "Validation".

    Examples
    --------
    >>> # Place the downloaded folders in any folder and pass the path to the constructor:
    >>> ds = FLARE2022(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 2100
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 110)
    >>> print(ds.mask(ds.ids[25]).shape)
    # (512, 512, 104)

    References
    ----------
    .. [1] Ma, Jun, et al. "Fast and Low-GPU-memory abdomen CT organ segmentation: The FLARE challenge."
    Medical Image Analysis 82 (2022): 102616.
    """

    @property
    def ids(self):
        result = set()

        # 50 Training Labeled cases
        archive = self.root / 'Training' / 'FLARE22_LabeledCase50' / 'images.zip'
        with ZipFile(archive) as zf:
            for file in zf.namelist():
                result.add(f"TL{file.split('_')[-2]}")

        # 2000 Training Unlabeled cases
        for archive in (self.root / 'Training').glob('*.zip'):
            with ZipFile(archive) as zf:
                for file in zf.namelist():
                    if not file.endswith('.nii.gz'):
                        continue

                    file = Path(file)
                    result.add(f"TU{file.name.split('_')[-2]}")

        # 50 Validation Unlabeled cases
        for file in (self.root / 'Validation').glob('*'):
            if not file.name.endswith('.nii.gz'):
                continue

            result.add(f"VU{file.name.split('_')[-2]}")

        return sorted(result)

    def _file(self, i):
        # 50 Training Labeled cases
        if i.startswith('TL'):
            archive = self.root / 'Training' / 'FLARE22_LabeledCase50' / 'images.zip'
            with ZipFile(archive) as zf:
                for file in zf.namelist():
                    if i[2:] in file:
                        return zipfile.Path(archive, file)

        # 2000 Training Unlabeled cases
        for archive in (self.root / 'Training').glob('*.zip'):
            with ZipFile(archive) as zf:
                for file in zf.namelist():
                    if i[2:] in file:
                        return zipfile.Path(archive, file)

        # 50 Validation Unlabeled cases
        if i.startswith('VU'):
            file = self.root / 'Validation' / f'FLARETs_{i[2:]}_0000.nii.gz'
            return file

        raise ValueError(f'Id "{i}" not found')

    @field
    def image(self, i) -> np.ndarray:
        with self._file(i).open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return np.asarray(image.dataobj)

    @field
    def affine(self, i) -> np.ndarray:
        """The 4x4 matrix that gives the image's spatial orientation"""
        with self._file(i).open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return image.affine

    @field
    def mask(self, i) -> Union[np.ndarray, None]:
        if not i.startswith('TL'):
            return None

        archive = self.root / 'Training' / 'FLARE22_LabeledCase50' / 'labels.zip'
        with ZipFile(archive) as zf:
            for file in zf.namelist():
                if i[2:] in file:
                    with zipfile.Path(archive, file).open('rb') as opened:
                        with gzip.GzipFile(fileobj=opened) as nii:
                            nii = nibabel.FileHolder(fileobj=nii)
                            mask = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                            return np.asarray(mask.dataobj)



import datetime
import os
from typing import List, Tuple, Union

import numpy as np
import pylidc as pl
from dicom_csv import (
    Series,
    expand_volumetric,
    get_common_tag,
    get_orientation_matrix,
    get_tag,
    order_series,
    stack_images,
)
from pylidc.utils import consensus
from scipy import stats

from ..internals import Dataset, field, licenses, register
from ..utils import PathOrStr, get_series_date
from .nodules import get_nodule
from .typing import LIDCNodule


@register(
    body_region='Chest',
    license=licenses.CC_BY_30,
    link='https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=1966254',
    modality='CT',
    prep_data_size='71,2G',
    raw_data_size='126G',
    task='Lung nodules segmentation',
)
class LIDC(Dataset):
    """
    The (L)ung (I)mage (D)atabase (C)onsortium image collection (LIDC-IDRI) [1]_
    consists of diagnostic and lung cancer screening thoracic computed tomography (CT) scans
    with marked-up annotated lesions and lung nodules segmentation task.
    Scans contains multiple expert annotations.

    Number of CT scans: 1018.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Follow the download instructions at https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=1966254.

    Then, the folder with raw downloaded data should contain folder `LIDC-IDRI`,
    which contains folders `LIDC-IDRI-*`.

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = LIDC(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 1018
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 194)
    >>> print(ds.cancer(ds.ids[0]).shape)
    # (512, 512, 194)

    References
    ----------
    .. [1] Armato III, McLennan, et al. "The lung image database consortium (lidc) and image database
    resource initiative (idri): a completed reference database of lung nodules on ct scans."
    Medical physics 38(2) (2011): 915–931.
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3041807/
    """

    def __init__(self, root: PathOrStr):
        super().__init__(root)
        self._check_config()

    def _check_config(self):
        pylidc_config_start = '[dicom]\npath = '
        if os.path.exists(os.path.expanduser('~/.pylidcrc')):
            with open(os.path.expanduser('~/.pylidcrc'), 'r') as config_file:
                content = config_file.read()
            if content == f'{pylidc_config_start}{self.root}':
                return

        # save _root path to ~/.pylidcrc file for pylidc
        with open(os.path.expanduser('~/.pylidcrc'), 'w') as config_file:
            config_file.write(f'{pylidc_config_start}{self.root}')

    @property
    def ids(self):
        result = [scan.series_instance_uid for scan in pl.query(pl.Scan).all()]
        return tuple(sorted(result))

    def _scan(self, i) -> pl.Scan:
        _id = i.split('_')[-1]
        return pl.query(pl.Scan).filter(pl.Scan.series_instance_uid == _id).first()

    def _series(self, i) -> Series:
        series = expand_volumetric(self._scan(i).load_all_dicom_images(verbose=False))
        series = order_series(series)
        return series

    def _shape(self, i) -> Tuple[int, int, int]:
        return stack_images(self._series(i), -1).shape

    @field
    def image(self, i) -> np.ndarray:
        return self._scan(i).to_volume(verbose=False)

    @field
    def study_uid(self, i) -> str:
        return self._scan(i).study_instance_uid

    @field
    def series_uid(self, i) -> str:
        return self._scan(i).series_instance_uid

    @field
    def patient_id(self, i) -> str:
        return self._scan(i).patient_id

    @field
    def sop_uids(self, i) -> List[str]:
        return [str(get_tag(i, 'SOPInstanceUID')) for i in self._series(i)]

    @field
    def pixel_spacing(self, i) -> List[float]:
        spacing = self._scan(i).pixel_spacing
        return [spacing, spacing]

    @field
    def slice_locations(self, i) -> np.ndarray:
        return self._scan(i).slice_zvals

    # @field
    def spacing(self, i) -> Tuple[float, float, float]:
        """
        Volumetric spacing of the image.
        The maximum relative difference in `slice_locations` < 1e-3
        (except 4 images listed below),
        so we allow ourselves to use the common spacing for the whole 3D image.

        Note
        ----
        The `slice_locations` attribute typically (but not always!) has the constant step.
        In LIDC dataset, only 4 images have difference in `slice_locations` > 1e-3:
            1.3.6.1.4.1.14519.5.2.1.6279.6001.526570782606728516388531252230
            1.3.6.1.4.1.14519.5.2.1.6279.6001.329334252028672866365623335798
            1.3.6.1.4.1.14519.5.2.1.6279.6001.245181799370098278918756923992
            1.3.6.1.4.1.14519.5.2.1.6279.6001.103115201714075993579787468219
        And these differences appear in the maximum of 3 slices.
        Therefore, we consider their impact negligible.
        """
        return (*self.pixel_spacing(i), stats.mode(np.diff(self.slice_locations(i)))[0].item())

    @field
    def contrast_used(self, i) -> bool:
        """If the DICOM file for the scan had any Contrast tag, this is marked as `True`."""
        return self._scan(i).contrast_used

    @field
    def is_from_initial(self, i) -> bool:
        """
        Indicates whether or not this PatientID was tagged as
        part of the initial 399 release.
        """
        return self._scan(i).is_from_initial

    @field
    def orientation_matrix(self, i) -> np.ndarray:
        return get_orientation_matrix(self._series(i))

    @field
    def sex(self, i) -> Union[str, None]:
        return get_common_tag(self._series(i), 'PatientSex', default=None)

    @field
    def age(self, i) -> Union[str, None]:
        return get_common_tag(self._series(i), 'PatientAge', default=None)

    @field
    def conv_kernel(self, i) -> Union[str, None]:
        return get_common_tag(self._series(i), 'ConvolutionKernel', default=None)

    @field
    def kvp(self, i) -> Union[str, None]:
        return get_common_tag(self._series(i), 'KVP', default=None)

    @field
    def tube_current(self, i) -> Union[str, None]:
        return get_common_tag(self._series(i), 'XRayTubeCurrent', default=None)

    @field
    def study_date(self, i) -> Union[datetime.date, None]:
        return get_series_date(self._series(i))

    @field
    def accession_number(self, i) -> Union[str, None]:
        return get_common_tag(self._series(i), 'AccessionNumber', default=None)

    @field
    def nodules(self, i) -> List[List[LIDCNodule]]:
        nodules = []
        for anns in self._scan(i).cluster_annotations():
            nodule_annotations = []
            for ann in anns:
                nodule_annotations.append(get_nodule(ann))
            nodules.append(nodule_annotations)
        return nodules

    @field
    def nodules_masks(self, i) -> List[List[np.ndarray]]:
        nodules = []
        for anns in self._scan(i).cluster_annotations():
            nodule_annotations = []
            for ann in anns:
                nodule_annotations.append(ann.boolean_mask())
            nodules.append(nodule_annotations)
        return nodules

    @field
    def cancer(self, i) -> np.ndarray:
        cancer = np.zeros(self._shape(i), dtype=bool)
        for anns in self._scan(i).cluster_annotations():
            cancer |= consensus(anns, pad=np.inf)[0]

        return cancer



import gzip
import json
import zipfile
from pathlib import Path
from typing import Dict, Tuple, Union
from zipfile import ZipFile

import nibabel
import numpy as np

from .internals import Dataset, field, licenses, register


@register(
    body_region=('Thorax', 'Abdomen'),
    modality='CT',
    task='Vertebrae Segmentation',
    link='https://osf.io/4skx2/',
    raw_data_size='97G',
    license=licenses.CC_BYSA_40,
)
class VerSe(Dataset):
    """
    A Vertebral Segmentation Dataset with Fracture Grading [1]_

    The dataset was used in the MICCAI-2019 and MICCAI-2020 Vertebrae Segmentation Challenges.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Download links:
        2019: https://osf.io/jtfa5/
        2020: https://osf.io/4skx2/

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = VerSe(root='/path/to/archives/root')
    >>> print(len(ds.ids))
    # 374
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 214)

    References
    ----------
    .. [1] Löffler MT, Sekuboyina A, Jacob A, et al. A Vertebral Segmentation Dataset with Fracture Grading.
       Radiol Artif Intell. 2020;2(4):e190138. Published 2020 Jul 29. doi:10.1148/ryai.2020190138
    """

    @property
    def ids(self):
        result = set()
        for archive in self.root.glob('*.zip'):
            with ZipFile(archive) as zf:
                for file in zf.namelist():
                    if '/rawdata/' not in file:
                        continue

                    file = Path(file)
                    patient = file.parent.name[4:]
                    name = file.name
                    if 'split' in name:
                        i = name.split('split')[1][1:]
                        i = i.split('_')[0]
                    else:
                        i = patient

                    assert i not in result, i
                    result.add(i)

        return sorted(result)

    def _file(self, i):
        for archive in self.root.glob('*.zip'):
            with ZipFile(archive) as zf:
                for file in zf.namelist():
                    if '/rawdata/' in file and i in file:
                        return zipfile.Path(archive, file)

        raise ValueError(f'Id "{i}" not found')

    @field
    def image(self, i) -> np.ndarray:
        with self._file(i).open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                # most ct scans are integer-valued, this will help us improve compression rates
                #  (instead of using `image.get_fdata()`)
                return np.asarray(image.dataobj)

    @field
    def affine(self, i) -> np.ndarray:
        """The 4x4 matrix that gives the image's spatial orientation"""
        with self._file(i).open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return image.affine

    @field
    def split(self, i) -> str:
        """The split in which this entry is contained: training, validate, test"""
        # it's ugly, but it gets the job done (;
        return self._file(i).parent.parent.parent.name.split('_')[-1].split('9')[-1]

    @field
    def patient(self, i) -> str:
        """The unique patient id"""
        return self._file(i).parent.name[4:]

    @field
    def year(self, i) -> int:
        """The year in which this entry was published: 2019, 2020"""
        year = self._file(i).parent.parent.parent.name
        if year.startswith('dataset-verse'):
            assert '19' in year
            return 2019
        return 2020

    def _derivatives(self, i):
        file = self._file(i)
        return file.parent.parent.parent / 'derivatives' / file.parent.name

    @field
    def centers(self, i) -> Dict[str, Tuple[int, int, int]]:
        """Vertebrae centers in format {label: [x, y, z]}"""
        ann = [f for f in self._derivatives(i).iterdir() if f.name.endswith('.json') and i in f.name]
        if not ann:
            return {}
        assert len(ann) == 1
        (ann,) = ann

        with ann.open() as file:
            ann = json.load(file)

        return {k['label']: (k['X'], k['Y'], k['Z']) for k in ann[1:]}

    @field
    def masks(self, i) -> Union[np.ndarray, None]:
        """Vertebrae masks"""
        ann = [f for f in self._derivatives(i).iterdir() if f.name.endswith('.nii.gz') and i in f.name]
        if not ann:
            return
        assert len(ann) == 1
        (ann,) = ann

        with ann.open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                mask = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return mask.get_fdata().astype(np.uint8)



from functools import cached_property

import deli
import nibabel
import numpy as np

from .internals import Dataset, register


@register(
    body_region=('Abdomen', 'Thorax'),
    link='https://nihcc.app.box.com/v/DeepLesion',
    modality='CT',
    prep_data_size='259G',
    raw_data_size='259G',
    task=('Localisation', 'Detection', 'Classification'),
)
class DeepLesion(Dataset):
    """
    DeepLesion is composed of 33,688 bookmarked radiology images from
    10,825 studies of 4,477 unique patients. For every bookmarked image, a bound-
    ing box is created to cover the target lesion based on its measured diameters [1].

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing `DL_info.csv` file and a subfolder `Images_nifti` with 20094 nii.gz files.

    Notes
    -----
    Dataset is available at https://nihcc.app.box.com/v/DeepLesion

    To download the data we recommend using a Python script provided by the authors `batch_download_zips.py`.
    Once you download the data and unarchive all 56 zip archives, you should run `DL_save_nifti.py`
    provided by the authors to convert 2D PNGs into 20094 nii.gz files.

    Example
    --------
    >>> ds = DeepLesion(root='/path/to/folder')
    >>> print(len(ds.ids))
    # 20094

    References
    ----------
    .. [1] Yan, Ke, Xiaosong Wang, Le Lu, and Ronald M. Summers. "Deeplesion: Automated deep mining,
      categorization and detection of significant radiology image findings using large-scale clinical
      lesion annotations." arXiv preprint arXiv:1710.01766 (2017).

    """

    @property
    def ids(self):
        return tuple(sorted(file.name.replace('.nii.gz', '') for file in (self.root / 'Images_nifti').glob('*.nii.gz')))

    def _image_file(self, i):
        return nibabel.load(self.root / 'Images_nifti' / f'{i}.nii.gz')

    @cached_property
    def _metadata(self):
        df = deli.load(self.root / 'DL_info.csv')

        cols_to_transform = [
            'Measurement_coordinates',
            'Bounding_boxes',
            'Lesion_diameters_Pixel_',
            'Normalized_lesion_location',
        ]
        for col in cols_to_transform:
            df[col] = df[col].apply(lambda x: list(map(float, x.split(','))))

        df['Slice_range_list'] = df['Slice_range'].apply(lambda x: list(map(int, x.split(','))))

        def get_ids(x):
            patient_study_series = '_'.join(x.File_name.split('_')[:3])
            slice_range_list = list(map(str, x.Slice_range_list))
            slice_range_list = [num.zfill(3) for num in slice_range_list]
            slice_range_list = '-'.join(slice_range_list)
            return f'{patient_study_series}_{slice_range_list}'

        df['ids'] = df.apply(get_ids, axis=1)
        return df

    def _row(self, i):
        # funny story, f-string does not work for pandas.query,
        # @ syntax does not work for linter, use # noqa
        return self._metadata.query('ids==@i')

    def patient_id(self, i):
        patient, study, series = map(int, i.split('_')[:3])
        return patient

    def study_id(self, i):
        patient, study, series = map(int, i.split('_')[:3])
        return study

    def series_id(self, i):
        patient, study, series = map(int, i.split('_')[:3])
        return series

    def sex(self, i):
        return self._row(i).Patient_gender.iloc[0]

    def age(self, i):
        """Patient Age might be different for different studies (dataset contains longitudinal records)."""
        return self._row(i).Patient_age.iloc[0]

    def ct_window(self, i):
        """CT window extracted from DICOMs. Recall, that it is min-max values for windowing, not width-level."""
        return self._row(i).DICOM_windows.iloc[0]

    def affine(self, i):
        return self._image_file(i).affine

    def spacing(self, i):
        return tuple(self._image_file(i).header['pixdim'][1:4])

    def image(self, i):
        """Some 3D volumes are stored as separate subvolumes, e.g. ds.ids[15000] and ds.ids[15001]."""
        return np.asarray(self._image_file(i).dataobj)

    def train_val_test(self, i):
        """Authors' defined randomly generated patient-level data split, train=1, validation=2, test=3,
        70/15/15 ratio."""
        return int(self._row(i).Train_Val_Test.iloc[0])

    def lesion_position(self, i):
        """Lesion measurements as it appear in DL_info.csv, for details see
        https://nihcc.app.box.com/v/DeepLesion/file/306056134060 ."""
        position = self._row(i)[
            [
                'Slice_range_list',
                'Key_slice_index',
                'Measurement_coordinates',
                'Bounding_boxes',
                'Lesion_diameters_Pixel_',
                'Normalized_lesion_location',
            ]
        ].to_dict('list')
        position['Slice_range_list'] = position['Slice_range_list'][0]
        return position

    def mask(self, i):
        """Mask of provided bounding boxes. Recall that bboxes annotation
        is very coarse, it only covers a single 2D slice."""
        mask = np.zeros_like(self.image(i))
        lesion_position = self.lesion_position(i)
        min_index = lesion_position['Slice_range_list'][0]
        for i, slice_index in enumerate(lesion_position['Key_slice_index']):
            image_index = slice_index - min_index
            top_left_x, top_left_y, bot_right_x, bot_right_y = lesion_position['Bounding_boxes'][i]
            mask[
                int(np.floor(top_left_y)) : int(np.ceil(bot_right_y)),
                int(np.floor(top_left_x)) : int(np.ceil(bot_right_x)),
                image_index,
            ] = 1
        return mask



import gzip
from contextlib import suppress
from pathlib import Path
from zipfile import ZipFile

import nibabel
import numpy as np
import pandas as pd

from ..internals import Dataset, field, licenses, register
from ..utils import PathOrStr, open_nii_gz_file, unpack
from .utils import ARCHIVE_ROOT, add_labels, add_masks


@register(
    body_region=('Head', 'Thorax', 'Abdomen', 'Pelvis', 'Legs'),
    license=licenses.CC_BY_40,
    link='https://zenodo.org/record/6802614#.Y6M2MxXP1D8',
    modality='CT',
    raw_data_size='35G',
    prep_data_size='35G',
    task='Supervised anatomical structures segmentation',
)
class Totalsegmentator(Dataset):
    """
    In 1204 CT images we segmented 104 anatomical structures (27 organs, 59 bones, 10 muscles, 8 vessels)
    covering a majority of relevant classes for most use cases.

    The CT images were randomly sampled from clinical routine, thus representing a real world dataset which
    generalizes to clinical application.

    The dataset contains a wide range of different pathologies, scanners, sequences and institutions. [1]

    Parameters
    ----------
    root : str, Path, optional
        absolute path to the downloaded archive.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Download link: https://zenodo.org/record/6802614/files/Totalsegmentator_dataset.zip

    Examples
    --------
    >>> # Download the archive to any folder and pass the path to the constructor:
    >>> ds = Totalsegmentator(root='/path/to/the/downloaded/archive')
    >>> print(len(ds.ids))
    # 1204
    >>> print(ds.image(ds.ids[0]).shape)
    # (294, 192, 179)
    >>> print(ds.aorta(ds.ids[25]).shape)
    # (320, 320, 145)

    References
    ----------
    .. [1] Jakob Wasserthal (2022) Dataset with segmentations of 104 important anatomical structures in 1204 CT images.
    Available at: https://zenodo.org/record/6802614#.Y6M2MxXP1D8
    """

    add_masks(locals())
    add_labels(locals())

    def __init__(self, root: PathOrStr):
        root = Path(root)
        if root.is_dir():
            if root / ARCHIVE_ROOT in list(root.iterdir()):
                root = root / ARCHIVE_ROOT

        file = 'meta.csv'
        with unpack(root, file, ARCHIVE_ROOT, '.zip') as (unpacked, _):
            self._meta = pd.read_csv(unpacked, sep=';')

        super().__init__(root)

    @property
    def ids(self):
        if self.root.is_dir():
            return sorted({x.name for x in self.root.iterdir() if x.name != 'meta.csv'})
        else:
            with ZipFile(self.root) as zf:
                parsed_namelist = [x.strip('/').split('/') for x in zf.namelist()]
                return sorted({x[-1] for x in parsed_namelist if len(x) == 2 and x[-1] != 'meta.csv'})

    @field
    def image(self, i):
        file = f'{i}/ct.nii.gz'

        with suppress(gzip.BadGzipFile):
            with unpack(self.root, file, ARCHIVE_ROOT, '.zip') as (unpacked, is_unpacked):
                if is_unpacked:
                    return np.asarray(nibabel.load(unpacked).dataobj)
                else:
                    with open_nii_gz_file(unpacked) as image:
                        return np.asarray(image.dataobj)

    @field
    def affine(self, i):
        """The 4x4 matrix that gives the image's spatial orientation"""
        file = f'{i}/ct.nii.gz'

        with unpack(self.root, file, ARCHIVE_ROOT, '.zip') as (unpacked, is_unpacked):
            if is_unpacked:
                return nibabel.load(unpacked).affine
            else:
                with open_nii_gz_file(unpacked) as image:
                    return image.affine



import deli
import numpy as np
import pydicom
from dicom_csv import (
    Plane,
    drop_duplicated_slices,
    expand_volumetric,
    get_common_tag,
    get_orientation_matrix,
    get_pixel_spacing,
    get_slice_locations,
    get_slices_plane,
    get_tag,
    order_series,
    stack_images,
)
from tqdm.auto import tqdm

from .internals import Dataset, field, licenses, register
from .utils import get_series_date


@register(
    body_region='Thorax',
    license=licenses.CC_BY_30,
    link='https://wiki.cancerimagingarchive.net/display/NLST/National+Lung+Screening+Trial',
    modality='CT',
    prep_data_size=None,  # TODO: should be measured...
    raw_data_size=None,  # TODO: should be measured...
    task=None,
)
class NLST(Dataset):
    """

        Dataset with low-dose CT scans of 26,254 patients acquired during National Lung Screening Trial.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder (usually called NLST) containing the patient subfolders (like 101426).
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Follow the download instructions at
    https://wiki.cancerimagingarchive.net/display/NLST/National+Lung+Screening+Trial.
    The dicoms should be placed under the following folders' structure:
        <...>/<NLST-root>/<patiend_id>/<study_uid>/<date>/<series_uid>/*.dcm

    Examples
    --------
    >>> ds = NLST(root='/path/to/NLST/')
    >>> print(len(ds.ids))
     ...
    >>> print(ds.image(ds.ids[0]).shape)
     ...
    >>> print(ds.mask(ds.ids[80]).shape)
     ...

    References
    ----------
    """

    @property
    def ids(self):
        ids = []
        for path in tqdm(list(self.root.iterdir())):
            series_uid2num_slices = {p.stem: int(deli.load(p)['Total'][5]) for p in path.glob('*/*/*.json')}
            ids.append(max(series_uid2num_slices, key=series_uid2num_slices.get))

        return ids

    def _series(self, i):
        (folder,) = self.root.glob(f'**/{i}')
        series = list(map(pydicom.dcmread, folder.iterdir()))
        series = expand_volumetric(series)
        assert get_common_tag(series, 'Modality') == 'CT'
        assert get_slices_plane(series) == Plane.Axial
        series = drop_duplicated_slices(series)
        series = order_series(series, decreasing=False)
        return series

    @field
    def image(self, i):
        return np.moveaxis(stack_images(self._series(i), -1).astype(np.int16), 0, 1)

    @field
    def study_uid(self, i):
        return get_common_tag(self._series(i), 'StudyInstanceUID')

    @field
    def series_uid(self, i):
        return get_common_tag(self._series(i), 'SeriesInstanceUID')

    @field
    def sop_uids(self, i):
        return [str(get_tag(i, 'SOPInstanceUID')) for i in self._series(i)]

    @field
    def pixel_spacing(self, i):
        return get_pixel_spacing(self, i).tolist()

    @field
    def slice_locations(self, i):
        return get_slice_locations(self, i)

    @field
    def orientation_matrix(self, i):
        return get_orientation_matrix(self, i)

    @field
    def conv_kernel(self, i):
        return get_common_tag(self._series(i), 'ConvolutionKernel', default=None)

    @field
    def kvp(self, i):
        return get_common_tag(self._series(i), 'KVP', default=None)

    @field
    def patient_id(self, i):
        return get_common_tag(self._series(i), 'PatientID', default=None)

    @field
    def study_date(self, i):
        return get_series_date(self._series(i))

    @field
    def accession_number(self, i):
        return get_common_tag(self._series(i), 'AccessionNumber', default=None)



from functools import cached_property
from typing import Dict, Tuple

import numpy as np
import SimpleITK
from connectome import Transform
from deli import load
from imops import crop_to_box

from .internals import Dataset, field, licenses, register
from .utils import mask_to_box


@register(
    body_region='Head',
    license=licenses.CC_BYNC_40,
    link='https://github.com/cwwang1979/CL-detection2023/',
    modality='X-ray',
    prep_data_size='1.8G',
    raw_data_size='1.5G',
    task='Keypoint detection',
)
class CLDetection2023(Dataset):
    """
    The data for the "Cephalometric Landmark Detection in Lateral X-ray Images" Challenge,
    held with the MICCAI-2023 conference.

    Notes
    -----
    The data can only be obtained by contacting the organizers by email.
    See the [challenge home page](https://cl-detection2023.grand-challenge.org/) for details.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded and unarchived data.
        If not provided, the cache is assumed to be already populated.

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = CLDetection2023(root='/path/to/data/root/folder')
    >>> print(len(ds.ids))
    # 400
    >>> print(ds.image(ds.ids[0]).shape)
    # (2400, 1935)
    """

    @cached_property
    def _images(self):
        return SimpleITK.GetArrayFromImage(SimpleITK.ReadImage(self.root / 'train_stack.mha'))

    @cached_property
    def _points(self):
        return load(self.root / 'train-gt.json')['points']

    @property
    def ids(self):
        return tuple(map(str, range(1, len(self._images) + 1)))

    @field
    def image(self, i) -> np.ndarray:
        i = int(i)
        return self._images[i - 1]

    @field
    def points(self, i) -> Dict[str, np.ndarray]:
        i = int(i)
        return {x['name']: np.array(x['point'][:2]) for x in self._points if x['point'][-1] == i}

    @field
    def spacing(self, i) -> Tuple[float, float]:
        i = int(i)
        (scale,) = {x['scale'] for x in self._points if x['point'][-1] == i}
        scale = float(scale)
        return scale, scale


class CropPadding(Transform):
    __inherit__ = 'spacing'

    def _box(image):
        return mask_to_box(image[..., 0] != 0)

    def image(image, _box):
        return crop_to_box(image[..., 0], _box)

    def points(points, _box):
        return {k: v - _box[0] for k, v in points.items()}


class FlipPoints(Transform):
    __inherit__ = True

    def points(points):
        return {name: pt[::-1] for name, pt in points.items()}



from typing import Sequence, Union

import numpy as np
from connectome import Transform
from imops import zoom

from ..utils import Numeric


class CanonicalCTOrientation(Transform):
    __exclude__ = ('nodules', 'nodules_masks')

    def image(image):
        return image[..., ::-1]

    def cancer(cancer):
        return cancer[..., ::-1]


class Rescale(Transform):
    __exclude__ = ('pixel_spacing', 'slice_locations', 'voxel_spacing', 'orientation_matrix')

    _new_spacing: Union[Sequence[Numeric], Numeric]
    _order: int = 1

    def _spacing(spacing, _new_spacing):
        _new_spacing = np.broadcast_to(_new_spacing, len(spacing)).copy()
        _new_spacing[np.isnan(_new_spacing)] = np.array(spacing)[np.isnan(_new_spacing)]
        return tuple(_new_spacing.tolist())

    def _scale_factor(spacing, _spacing):
        return np.float32(spacing) / np.float32(_spacing)

    def spacing(_spacing):
        return _spacing

    def image(image, _scale_factor, _order):
        return zoom(image.astype(np.float32), _scale_factor, order=_order)

    def cancer(cancer, _scale_factor, _order):
        return zoom(cancer.astype(np.float32), _scale_factor, order=_order) > 0.5



import gzip
from typing import Union

import nibabel
import numpy as np

from .internals import Dataset, field, register


@register(
    body_region='Thorax',
    modality='CT',
    task='COVID-19 Segmentation',
    link='https://mosmed.ai/en/datasets/covid191110/',
    raw_data_size='21G',
)
class MoscowCovid1110(Dataset):
    """
    The Moscow Radiology COVID-19 dataset.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded files.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Download links:
    https://mosmed.ai/en/datasets/covid191110/

    Examples
    --------
    >>> # Place the downloaded files in any folder and pass the path to the constructor:
    >>> ds = MoscowCovid1110(root='/path/to/files/root')
    >>> print(len(ds.ids))
    # 1110
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 43)
    """

    @property
    def ids(self):
        return sorted({f.name[:-7] for f in self.root.glob('CT-*/*')})

    def _file(self, i):
        return next(self.root.glob(f'CT-*/{i}.nii.gz'))

    @field
    def image(self, i) -> np.ndarray:
        with self._file(i).open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                # most ct scans are integer-valued, this will help us improve compression rates
                #  (instead of using `image.get_fdata()`)
                return np.asarray(image.dataobj)

    @field
    def affine(self, i) -> np.ndarray:
        with self._file(i).open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return image.affine

    @field
    def label(self, i) -> str:
        return self._file(i).parent.name[3:]

    @field
    def mask(self, i) -> Union[np.ndarray, None]:
        path = self.root / 'masks' / f'{i}_mask.nii.gz'
        if not path.exists():
            return

        with path.open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return np.asarray(image.dataobj) > 0.5



import plistlib
import warnings
from ast import literal_eval
from enum import IntEnum
from functools import lru_cache
from pathlib import Path
from typing import NamedTuple, Union

import numpy as np
import pandas as pd
import pydicom
from connectome import Transform
from dicom_csv import (
    drop_duplicated_instances,
    drop_duplicated_slices,
    expand_volumetric,
    get_orientation_matrix,
    get_pixel_spacing,
    get_slice_locations,
    order_series,
    stack_images,
)
from skimage.draw import polygon

from .internals import Dataset, field, licenses, register


class CoCaClasses(IntEnum):
    LAD = 1
    LCX = 2
    RCA = 3
    LCA = 4


class Calcification(NamedTuple):
    label: str
    contour_px: np.ndarray
    contour_mm: np.ndarray
    center: np.ndarray
    area: float
    hu_min: float
    hu_max: float
    hu_dev: float
    hu_mean: float
    length: float
    total: int


@register(
    body_region=('Coronary', 'Chest'),
    license=licenses.StanfordDSResearch,
    link='https://stanfordaimi.azurewebsites.net/datasets/e8ca74dc-8dd4-4340-815a-60b41f6cb2aa',
    modality='CT',
    prep_data_size=None,  # TODO: should be measured...
    raw_data_size='28G',
    task='Coronary Calcium Segmentation',
)
class StanfordCoCa(Dataset):
    """
    A Stanford AIMI's Co(ronary) Ca(lcium) dataset.


    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.


    Notes
    -----
    Follow the download instructions at
    https://stanfordaimi.azurewebsites.net/datasets/e8ca74dc-8dd4-4340-815a-60b41f6cb2aa.
    You'll need to register and accept the terms of use. After that, copy the files from Azure:

    azcopy copy 'some-generated-access-link' /path/to/downloaded/data/ --recursive=true

    Then, the folder with raw downloaded data should contain two subfolders - a subset with gated coronary CT scans
    and corresponding coronary calcium segmentation masks (`Gated_release_final`)
    and a folder with the non-gated CT scans with corresponding coronary with coronary artery calcium scores
    (`deidentified_nongated`).

    The folder with gated data should have original structure:
        ./Gated_release_final/patient/0/folder-with-dcms/
        ./Gated_release_final/calcium_xml/0.xml
        ...

    The folder with nongated data should have original structure:
        ./deidentified_nongated/0/folder-with-dcms/
        ...

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = StanfordCoCa(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 971
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 57)

    """

    def _split(self, i):
        return i.split('-')[0]

    def _identifier(self, i):
        return i.split('-')[1]

    def _folder_with_images(self, i):
        split = self._split(i)
        if split == 'gated':
            return Path('Gated_release_final') / 'patient'
        if split == 'nongated':
            return 'deidentified_nongated'
        raise ValueError("Unknown split. Use 'gated' or 'nongated' options.")

    def _folder_with_annotations(self, i):
        split = self._split(i)
        if split == 'gated':
            return Path('Gated_release_final') / 'calcium_xml'
        if split == 'nongated':
            return None
        raise ValueError("Unknown split. Use 'gated' or 'nongated' options.")

    @property
    def ids(self):
        gated_ids = tuple(
            sorted('gated-' + x.name for x in (self.root / 'Gated_release_final' / 'patient').iterdir() if x.is_dir())
        )
        nongated_ids = tuple(
            sorted('nongated-' + x.name for x in (self.root / 'deidentified_nongated').iterdir() if x.is_dir())
        )

        return gated_ids + nongated_ids

    def _series(self, i):
        folder_with_dicoms = self.root / self._folder_with_images(i) / self._identifier(i)
        series = list(map(pydicom.dcmread, folder_with_dicoms.glob('*/*.dcm')))
        if not series:
            raise FileNotFoundError(f'No dicoms found at {folder_with_dicoms}')

        # series = sorted(series, key=lambda x: x.InstanceNumber)
        series = expand_volumetric(series)
        series = drop_duplicated_instances(series)

        original_num_slices = len(series)
        series = drop_duplicated_slices(series)
        if len(series) < original_num_slices:
            warnings.warn(f'Dropped duplicated slices for series {series[0]["StudyInstanceUID"]}.')

        series = order_series(series, decreasing=False)

        return series

    @field
    def image(self, i) -> np.ndarray:
        image = stack_images(self._series(i), -1).transpose((1, 0, 2)).astype(np.int16)
        return image

    def _image_meta(self, i):
        metas = [list(dict(s).values()) for s in self._series(i)]
        result = {}
        for meta_ in metas:
            for element in meta_:
                if element.keyword in ['PixelData']:
                    continue
                if element.keyword not in result:
                    result[element.keyword] = [element.value]
                elif result[element.keyword][-1] != element.value:
                    result[element.keyword].append(element.value)
        # turn elements that are the same across the series back from array
        result = {k: v[0] if len(v) == 1 else v for k, v in result.items()}
        return result

    @field
    def series_uid(self, i) -> str:
        return self._image_meta(i).get('SeriesInstanceUID', None)

    @field
    def study_uid(self, i) -> str:
        return self._image_meta(i).get('StudyInstanceUID', None)

    @field
    def pixel_spacing(self, i) -> Union[list, None]:
        series = self._series(i)
        return get_pixel_spacing(series).tolist() if series else None

    @field
    def slice_locations(self, i) -> Union[list, None]:
        series = self._series(i)
        return get_slice_locations(series) if series else None

    @field
    def orientation_matrix(self, i) -> Union[np.ndarray, None]:
        series = self._series(i)
        return get_orientation_matrix(series) if series else None

    def _raw_annotations(self, i):
        """Annotation as it is in xml"""
        folder = self._folder_with_annotations(i)
        if folder is None:
            warnings.warn("The used split doesn't contain segmentation masks.")
            return None

        try:
            with open(self.root / folder / f'{self._identifier(i)}.xml', 'rb') as fp:
                annotation = plistlib.load(fp)
                image_annotations = annotation['Images']

        except FileNotFoundError:
            warnings.warn(f'Missing annotation for id: {i}')
            return None

        return image_annotations

    @field
    def calcifications(self, i) -> Union[list, None]:
        """Returns list of Calcifications"""
        raw_annotations = self._raw_annotations(i)
        if raw_annotations is None:
            return None

        cacs = []
        for slice_annotation in raw_annotations:
            for roi in slice_annotation['ROIs']:
                if roi['Area'] > 0:
                    contour_px = np.array([literal_eval(x) for x in roi['Point_px']])
                    contour_mm = np.array([literal_eval(x) for x in roi['Point_mm']])
                    name, center, area = roi['Name'], roi['Center'], roi['Area']
                    hu_min, hu_max, hu_dev, hu_mean = roi['Min'], roi['Max'], roi['Dev'], roi['Mean']
                    total, length = roi['Total'], roi['Length']
                    cacs.append(
                        Calcification(
                            name, contour_px, contour_mm, center, area, hu_min, hu_max, hu_dev, hu_mean, total, length
                        )
                    )
        return cacs

    @lru_cache(None)
    def _scores(self, i):
        p = self.root / self._folder_with_images(i) / 'scores.xlsx'

        if not p.exists():
            return None

        return pd.read_excel(p, index_col=0)

    @field
    def score(self, i) -> Union[dict, None]:
        scores = self._scores(i)
        if scores is None:
            return None
        try:
            return scores.loc[i + 'A'].to_dict()
        except KeyError:
            warnings.warn(f'Missing scores for idx "{i}"')
            return None


class ContoursToMask(Transform):
    """Our implementation of transform for the contours. One can implement own logic based on this class"""

    __inherit__ = True
    _throw: bool = False
    _class_abbr: dict = {
        'Left Anterior Descending Artery': 'LAD',
        'Left Circumflex Artery': 'LCA',
        'Right Coronary Artery': 'RCA',
        'Left Coronary Artery': 'LCA',
    }

    def mask(id, calcifications, image, slice_locations, _class_abbr, _throw):
        if calcifications is None:
            return None

        shape = image.shape
        sl = slice_locations
        multiclass_mask = np.zeros(shape, np.uint8)
        try:
            for cac in calcifications:
                assert cac.label in _class_abbr, f'Unexpected class: {cac.label}'
                class_name = _class_abbr[cac.label]
                class_id = CoCaClasses[class_name].value
                slice_location = cac.contour_mm[0][-1]
                slice_id = np.argwhere(sl == slice_location).squeeze()
                assert slice_id, 'Slice where calcification is located is not presented in the series.'
                roi_contour = cac.contour_px
                slice_mask = np.zeros(shape[:2])
                xs, ys = polygon(*roi_contour.T, shape[:2])
                slice_mask[xs, ys] = True
                multiclass_mask[..., slice_id] = (class_id * slice_mask.astype(int)).astype(np.uint8)

        except AssertionError as e:
            if _throw:
                raise e
            else:
                warnings.warn(f"Mask preparation for idx {id} failed with: '{str(e)}'. Returning None")
                return None

        return multiclass_mask



from functools import partial
from typing import Dict

import highdicom
import numpy as np
from dicom_csv import get_orientation_matrix, get_slice_locations, get_voxel_spacing, stack_images
from imops import restore_crop
from more_itertools import locate

from .internals import Dataset, licenses, register
from .utils import series_from_dicom_folder


@register(
    body_region='Abdomen',
    license=licenses.CC_BY_40,
    link='https://wiki.cancerimagingarchive.net/pages/viewpage.action?'
    'pageId=89096268#89096268412b832037484784bd78caf58e052641',
    modality=('CT, SEG'),
    prep_data_size='11G',
    raw_data_size='11G',
    task=('Segmentation', 'Classification'),
)
class CRLM(Dataset):
    """
    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.


    Notes
    -----
    Download links:
    https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=89096268#89096268b2cc35fce0664a2b875b5ec675ba9446

    This collection consists of DICOM images and DICOM Segmentation Objects (DSOs)
    for 197 patients with Colorectal Liver Metastases (CRLM).
    Comprised of Original DICOM CTs and Segmentations for each subject.
    The segmentations include 'Liver', 'Liver_Remnant'
    (liver that will remain after surgery based on a preoperative CT plan),
    'Hepatic' and 'Portal' veins,
    and 'Tumor_x', where 'x' denotes the various tumor occurrences in the case

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = CRLM(root='/path/to/archives/root')
    >>> print(len(ds.ids))
    # 197
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 52)

    References
    ----------
    """

    @property
    def ids(self):
        return sorted(d.name for d in self.root.iterdir())

    def _folders(self, i):
        case = self.root / i
        folders = tuple({p.parent for p in case.glob('*/*/*/*.dcm')})
        return tuple(sorted(folders, key=lambda f: len(list(f.iterdir()))))

    def _series(self, i):
        return series_from_dicom_folder(self._folders(i)[1])

    def image(self, i):
        return stack_images(self._series(i))

    def mask(self, i) -> Dict[str, np.ndarray]:
        """Returns dict: {'liver': ..., 'hepatic': ..., 'tumor_x': ...}"""
        dicom_seg = highdicom.seg.segread(next(self._folders(i)[0].glob('*.dcm')))
        series = self._series(i)
        image_sops = [s.SOPInstanceUID for s in series]
        seg_sops = [sop_uid for _, _, sop_uid in dicom_seg.get_source_image_uids()]

        sops = [sop for sop in image_sops if sop in set(seg_sops).intersection(image_sops)]
        seg_box_start = list(locate(image_sops, lambda i: i == sops[0]))[0]
        seg_box_stop = list(locate(image_sops, lambda i: i == sops[-1]))[0]

        image = self.image(i)
        seg_box = np.asarray(((0, 0, seg_box_start), (*np.atleast_1d(image.shape[:-1]), seg_box_stop + 1)))

        raw_masks = np.swapaxes(
            dicom_seg.get_pixels_by_source_instance(
                sops,
                ignore_spatial_locations=True,
                segment_numbers=dicom_seg.get_segment_numbers(),
            ),
            -1,
            0,
        )
        masks = list(map(partial(restore_crop, box=seg_box, shape=image.shape), raw_masks))

        liver_mask = {'liver': masks[0].astype(bool)}
        # skip liver remnant
        veins = {'hepatic': masks[2].astype(bool), 'portal': masks[3].astype(bool)}
        tumors = {f'tumor_{i}': array.astype(bool) for i, array in enumerate(masks[4:])}

        return {**liver_mask, **veins, **tumors}

    def spacing(self, i):
        """Returns the voxel spacing along axes (x, y, z)."""
        return get_voxel_spacing(self._series(i))

    def slice_locations(self, i):
        return get_slice_locations(self._series(i))

    def affine(self, i):
        """Returns 4x4 matrix that gives the image's spatial orientation."""
        return get_orientation_matrix(self._series(i))



import nibabel as nb
import numpy as np
import pandas as pd

from .internals import Dataset, field, licenses, register


@register(
    body_region='Head',
    license=licenses.PhysioNet_RHD_150,
    link='https://physionet.org/content/ct-ich/1.3.1/',
    modality='CT',
    prep_data_size='661M',
    raw_data_size='2,8G',
    task='Intracranial hemorrhage segmentation',
)
class CT_ICH(Dataset):
    """
    (C)omputed (T)omography Images for (I)ntracranial (H)emorrhage Detection and (S)egmentation.

    This dataset contains 75 head CT scans including 36 scans for patients diagnosed with
    intracranial hemorrhage with the following types:
    Intraventricular, Intraparenchymal, Subarachnoid, Epidural and Subdural.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Data can be downloaded here: https://physionet.org/content/ct-ich/1.3.1/.
    Then, the folder with raw downloaded data should contain folders `ct_scans` and `masks` along with other files.

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = CT_ICH(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 75
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 39)
    >>> print(ds.mask(ds.ids[0]).shape)
    # (512, 512, 39)
    """

    @property
    def ids(self):
        result = [f'ct_ich_{uid:0=3d}' for uid in [*range(49, 59), *range(66, 131)]]
        return tuple(sorted(result))

    def _image_file(self, i):
        num_id = i.split('_')[-1]
        return nb.load(self.root / 'ct_scans' / f'{num_id}.nii')

    @field
    def image(self, i) -> np.ndarray:
        # most CT/MRI scans are integer-valued, this will help us improve compression rates
        return np.int16(self._image_file(i).get_fdata())

    @field
    def mask(self, i) -> np.ndarray:
        num_id = i.split('_')[-1]
        mask_path = self.root / 'masks' / f'{num_id}.nii'
        ct_scan_nifti = nb.load(mask_path)
        return ct_scan_nifti.get_fdata().astype(bool)

    @field
    def affine(self, i) -> np.ndarray:
        """The 4x4 matrix that gives the image's spatial orientation."""
        return self._image_file(i).affine

    def spacing(self, i):
        """Returns voxel spacing along axes (x, y, z)."""
        return tuple(self._image_file(i).header['pixdim'][1:4])

    @property
    def _patient_metadata(self):
        return pd.read_csv(self.root / 'Patient_demographics.csv', index_col='Patient Number')

    @property
    def _diagnosis_metadata(self):
        return pd.read_csv(self.root / 'hemorrhage_diagnosis_raw_ct.csv')

    def _row(self, i):
        patient_id = int(i.split('_')[-1])
        return self._patient_metadata.loc[patient_id]

    @field
    def age(self, i) -> float:
        return self._row(i)['Age\n(years)']

    @field
    def sex(self, i) -> str:
        return self._row(i)['Gender']

    @field
    def intraventricular_hemorrhage(self, i) -> bool:
        """Returns True if hemorrhage exists and its type is intraventricular."""
        num_id = int(i.split('_')[-1])
        return str(self._patient_metadata['Hemorrhage type based on the radiologists diagnosis '].loc[num_id]) != 'nan'

    @field
    def intraparenchymal_hemorrhage(self, i) -> bool:
        """Returns True if hemorrhage was diagnosed and its type is intraparenchymal."""
        num_id = int(i.split('_')[-1])
        return str(self._patient_metadata['Unnamed: 4'].loc[num_id]) != 'nan'

    @field
    def subarachnoid_hemorrhage(self, i) -> bool:
        """Returns True if hemorrhage was diagnosed and its type is subarachnoid."""
        num_id = int(i.split('_')[-1])
        return str(self._patient_metadata['Unnamed: 5'].loc[num_id]) != 'nan'

    @field
    def epidural_hemorrhage(self, i) -> bool:
        """Returns True if hemorrhage was diagnosed and its type is epidural."""
        num_id = int(i.split('_')[-1])
        return str(self._patient_metadata['Unnamed: 6'].loc[num_id]) != 'nan'

    @field
    def subdural_hemorrhage(self, i) -> bool:
        """Returns True if hemorrhage was diagnosed and its type is subdural."""
        num_id = int(i.split('_')[-1])
        return str(self._patient_metadata['Unnamed: 7'].loc[num_id]) != 'nan'

    @field
    def fracture(self, i) -> bool:
        """Returns True if skull fracture was diagnosed."""
        num_id = int(i.split('_')[-1])
        return str(self._patient_metadata['Fracture (yes 1/no 0)'].loc[num_id]) != 'nan'

    @field
    def notes(self, i) -> str:
        """Returns special notes if they exist."""
        num_id = int(i.split('_')[-1])
        result = str(self._patient_metadata['Note1'].loc[num_id])
        return result if result != 'nan' else None

    @field
    def hemorrhage_diagnosis_raw_metadata(self, i):
        num_id = int(i.split('_')[-1])
        return self._diagnosis_metadata[self._diagnosis_metadata['PatientNumber'] == num_id]



ANATOMICAL_STRUCTURES = [
    'adrenal_gland_left',
    'adrenal_gland_right',
    'aorta',
    'autochthon_left',
    'autochthon_right',
    'brain',
    'clavicula_left',
    'clavicula_right',
    'colon',
    'duodenum',
    'esophagus',
    'face',
    'femur_left',
    'femur_right',
    'gallbladder',
    'gluteus_maximus_left',
    'gluteus_maximus_right',
    'gluteus_medius_left',
    'gluteus_medius_right',
    'gluteus_minimus_left',
    'gluteus_minimus_right',
    'heart_atrium_left',
    'heart_atrium_right',
    'heart_myocardium',
    'heart_ventricle_left',
    'heart_ventricle_right',
    'hip_left',
    'hip_right',
    'humerus_left',
    'humerus_right',
    'iliac_artery_left',
    'iliac_artery_right',
    'iliac_vena_left',
    'iliac_vena_right',
    'iliopsoas_left',
    'iliopsoas_right',
    'inferior_vena_cava',
    'kidney_left',
    'kidney_right',
    'liver',
    'lung_lower_lobe_left',
    'lung_lower_lobe_right',
    'lung_middle_lobe_right',
    'lung_upper_lobe_left',
    'lung_upper_lobe_right',
    'pancreas',
    'portal_vein_and_splenic_vein',
    'pulmonary_artery',
    'rib_left_1',
    'rib_left_10',
    'rib_left_11',
    'rib_left_12',
    'rib_left_2',
    'rib_left_3',
    'rib_left_4',
    'rib_left_5',
    'rib_left_6',
    'rib_left_7',
    'rib_left_8',
    'rib_left_9',
    'rib_right_1',
    'rib_right_10',
    'rib_right_11',
    'rib_right_12',
    'rib_right_2',
    'rib_right_3',
    'rib_right_4',
    'rib_right_5',
    'rib_right_6',
    'rib_right_7',
    'rib_right_8',
    'rib_right_9',
    'sacrum',
    'scapula_left',
    'scapula_right',
    'small_bowel',
    'spleen',
    'stomach',
    'trachea',
    'urinary_bladder',
    'vertebrae_C1',
    'vertebrae_C2',
    'vertebrae_C3',
    'vertebrae_C4',
    'vertebrae_C5',
    'vertebrae_C6',
    'vertebrae_C7',
    'vertebrae_L1',
    'vertebrae_L2',
    'vertebrae_L3',
    'vertebrae_L4',
    'vertebrae_L5',
    'vertebrae_T1',
    'vertebrae_T10',
    'vertebrae_T11',
    'vertebrae_T12',
    'vertebrae_T2',
    'vertebrae_T3',
    'vertebrae_T4',
    'vertebrae_T5',
    'vertebrae_T6',
    'vertebrae_T7',
    'vertebrae_T8',
    'vertebrae_T9',
]

LABELS = ['age', 'gender', 'institute', 'study_type', 'split']



import gzip
import zipfile
from pathlib import Path
from zipfile import ZipFile

import nibabel as nb
import numpy as np

from .internals import Dataset, field, licenses, register


@register(
    body_region='Head',
    license=licenses.CC_BYNCND_40,
    link='https://www.humanconnectome.org/study/hcp-young-adult/document/1200-subjects-data-release',
    modality='MRI',
    prep_data_size='125G',
    raw_data_size='125G',
    task='Segmentation',
)
class HCP(Dataset):
    @property
    def ids(self):
        result = set()
        for archive in self.root.glob('*.zip'):
            with ZipFile(archive) as zf:
                for zipinfo in zf.infolist():
                    if zipinfo.is_dir():
                        continue
                    result.add(zipinfo.filename.split('/')[0])

        return tuple(sorted(result))

    def _file(self, i):
        for archive in self.root.glob('*.zip'):
            with ZipFile(archive) as zf:
                for zipinfo in zf.infolist():
                    if zipinfo.is_dir():
                        continue
                    file = Path(zipinfo.filename)
                    if (i in file.stem) and ('T1w_MPR1' in file.stem):
                        return zipfile.Path(str(archive), str(file))

    @field
    def image(self, i) -> np.ndarray:
        with self._file(i).open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nb.FileHolder(fileobj=nii)
                image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return np.int16(image.get_fdata())

    @field
    def affine(self, i) -> np.ndarray:
        with self._file(i).open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nb.FileHolder(fileobj=nii)
                image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return image.affine

    def spacing(self, i):
        with self._file(i).open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nb.FileHolder(fileobj=nii)
                image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return tuple(image.header['pixdim'][1:4])



import contextlib
import gzip
import zipfile
from pathlib import Path
from typing import Union
from zipfile import ZipFile

import nibabel as nb
import numpy as np
import pandas as pd

from .internals import Dataset, licenses, register


@register(
    body_region='Head',
    license=licenses.CC_BYNCSA_40,
    link='https://zenodo.org/record/6504722#.YsgwnNJByV4',
    modality=('MRI T1c', 'MRI T2hr'),
    prep_data_size='8,96G',
    raw_data_size='17G',
    task=('Segmentation', 'Classification', 'Domain Adaptation'),
)
class CrossMoDA(Dataset):
    """
    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Download links:
    2021 & 2022: https://zenodo.org/record/6504722#.YsgwnNJByV4

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = CrossMoDA(root='/path/to/archives/root')
    >>> print(len(ds.ids))
    # 484
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 214)

    References
    ----------
    """

    @property
    def ids(self):
        result = set()
        for archive in self.root.glob('*.zip'):
            with ZipFile(archive) as zf:
                for zipinfo in zf.infolist():
                    if zipinfo.is_dir():
                        continue

                    file = Path(zipinfo.filename)
                    assert file.stem not in result, file.stem

                    if 'Label' not in file.stem and file.suffix == '.gz':
                        result.add('_'.join(file.stem.split('_')[:-1]))
                    else:
                        continue

        return sorted(result)

    @property
    def train_source_df(self):
        return pd.read_csv(self.root / 'infos_source_training.csv', index_col='crossmoda_name')

    def _file(self, i):
        for archive in self.root.glob('*.zip'):
            with ZipFile(archive) as zf:
                for zipinfo in zf.infolist():
                    if i == '_'.join(Path(zipinfo.filename).stem.split('_')[:-1]) and 'Label' not in zipinfo.filename:
                        return zipfile.Path(archive, zipinfo.filename)

        raise ValueError(f'Id "{i}" not found')

    def image(self, i) -> Union[np.ndarray, None]:
        with open_nii_gz_file(self._file(i)) as nii_image:
            return np.asarray(nii_image.dataobj)

    def spacing(self, i):
        """Returns pixel spacing along axes (x, y, z)"""
        with open_nii_gz_file(self._file(i)) as nii_image:
            return tuple(nii_image.header['pixdim'][1:4])

    def affine(self, i):
        """The 4x4 matrix that gives the image's spatial orientation"""
        with open_nii_gz_file(self._file(i)) as nii_image:
            return nii_image.affine

    def split(self, i) -> str:
        """The split in which this entry is contained: training_source, training_target, validation"""
        file = self._file(i)
        idx = int(file.name.split('_')[2])
        dataset = file.name.split('_')[1]

        if dataset == 'ldn':
            if 1 <= idx < 106:
                return 'training_source'
            elif 106 <= idx < 211:
                return 'training_target'
            elif 211 <= idx < 243:
                return 'validation'

        elif dataset == 'etz':
            if 0 <= idx < 105:
                return 'training_source'
            elif 105 <= idx < 210:
                return 'training_target'
            elif 210 <= idx < 242:
                return 'validation'

        raise ValueError(f'Cannot find split for the file: {file}')

    def year(self, i) -> int:
        """The year in which this entry was published: 2021 or 2022"""
        return int(self._file(i).name[9:13])

    def masks(self, i):
        """Combined mask of schwannoma and cochlea (1 and 2 respectively)"""
        file = self._file(i)
        if 'T2' not in file.name:
            with open_nii_gz_file(file.parent / file.name.replace('ceT1', 'Label')) as nii_image:
                return nii_image.get_fdata().astype(np.uint8)

    def koos_grade(self, i):
        """VS Tumour characteristic according to Koos grading scale: [1..4] or (-1 - post operative)"""
        if self.split(i) == 'training_source':
            grade = self.train_source_df.loc[i, 'koos']
            return -1 if (grade == 'post-operative-london') else int(grade)


# TODO: sync with amid.utils
@contextlib.contextmanager
def open_nii_gz_file(file):
    with file.open('rb') as opened:
        with gzip.GzipFile(fileobj=opened) as nii:
            nii = nb.FileHolder(fileobj=nii)
            yield nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
